REFERENCE_BLOCK_IN_PEAK_CYCLE = 1000
CYCLES_IN_TESTING_DAY = 10  # 1 peak cycle + 9 non-peak cycles
TEMPO = 360


def get_epoch_containing_block(block: int, netuid: int, tempo: int = TEMPO) -> range:
    """
    Reimplementing the logic from subtensor's Rust function:
        pub fn blocks_until_next_epoch(netuid: u16, tempo: u16, block_number: u64) -> u64
    See https://github.com/opentensor/subtensor.

    See also: https://github.com/opentensor/bittensor/pull/2168/commits/9e8745447394669c03d9445373920f251630b6b8

    The beginning of an epoch is the first block when values like "dividends" are different
    (before an epoch they are constant for a full tempo).
    """
    assert tempo > 0

    interval = tempo + 1
    next_epoch = block + tempo - (block + netuid + 1) % interval

    if next_epoch == block:
        prev_epoch = next_epoch
        next_epoch = prev_epoch + interval
    else:
        prev_epoch = next_epoch - interval

    return range(prev_epoch, next_epoch)


def get_cycle_containing_block(block: int, netuid: int, tempo: int = TEMPO) -> range:
    """
    A cycle contains two epochs, starts on an even one. A cycle is the basic unit of passage of time in compute horde,
    and validators testing miners are synchronised to cycles.
    """
    very_first_epoch = get_epoch_containing_block(0, netuid, tempo=tempo)
    epoch_containing_block = get_epoch_containing_block(block, netuid, tempo=tempo)

    if ((epoch_containing_block.start - very_first_epoch.start) / (tempo + 1)) % 2:
        # that's the second epoch in this cycle
        first_epoch = range(
            epoch_containing_block.start - (tempo + 1), epoch_containing_block.stop - (tempo + 1)
        )
        second_epoch = epoch_containing_block
    else:
        first_epoch = epoch_containing_block
        second_epoch = range(
            epoch_containing_block.start + (tempo + 1), epoch_containing_block.stop + (tempo + 1)
        )

    return range(first_epoch.start, second_epoch.stop)


def get_peak_cycle(block: int, netuid: int, tempo: int = TEMPO) -> range:
    """Get the peak cycle of testing day containing block"""
    cycle_interval = 2 * (tempo + 1)
    testing_day_interval = cycle_interval * CYCLES_IN_TESTING_DAY
    reference_peak_cycle = get_cycle_containing_block(REFERENCE_BLOCK_IN_PEAK_CYCLE, netuid, tempo)
    excess_blocks = (block - reference_peak_cycle.start) % testing_day_interval
    peak_cycle = get_cycle_containing_block(block - excess_blocks, netuid, tempo)
    return peak_cycle



from dataclasses import dataclass
from datetime import timedelta

from compute_horde_core.executor_class import ExecutorClass


@dataclass
class ExecutorClassSpec:
    # mostly for user consumption, intended usage, ...
    description: str
    spin_up_time: int
    has_gpu: bool
    cpu_cores: int | None = None
    ram_gb: int | None = None
    storage_gb: int | None = None
    gpu_vram_gb: int | None = None
    docker_cached_images: list[str] | None = None
    # requirements which can't be easily standardized
    additional_requirements: str | None = None


EXECUTOR_CLASS = {
    ExecutorClass.spin_up_4min__gpu_24gb: ExecutorClassSpec(
        description="cost effective (started on demand), generic GPU machine",
        has_gpu=True,
        gpu_vram_gb=24,
        spin_up_time=int(timedelta(minutes=4).total_seconds()),
    ),
    ExecutorClass.always_on__gpu_24gb: ExecutorClassSpec(
        description="always on, generic GPU machine",
        has_gpu=True,
        gpu_vram_gb=24,
        spin_up_time=3,
    ),
    ExecutorClass.always_on__llm__a6000: ExecutorClassSpec(
        description="always on, NVIDIA RTX A6000 GPU machine for LLM prompts solving",
        has_gpu=True,
        gpu_vram_gb=48,
        spin_up_time=30,  # FIXME: temporary value for debugging
    ),
}

# This is the upper TTL for executors, after which executor pool kills an executor.
# TODO: TIMEOUTS - this should depend on the requested job timing instead, but capped at seconds left in current cycle
MAX_EXECUTOR_TIMEOUT = timedelta(minutes=20).total_seconds()

DEFAULT_EXECUTOR_CLASS = ExecutorClass.spin_up_4min__gpu_24gb
DEFAULT_LLM_EXECUTOR_CLASS = ExecutorClass.always_on__llm__a6000
DEFAULT_EXECUTOR_TIMEOUT = EXECUTOR_CLASS[DEFAULT_EXECUTOR_CLASS].spin_up_time



import random
from uuid import uuid4

from bittensor import Keypair
from compute_horde_core.executor_class import ExecutorClass
from django.utils import timezone

from compute_horde.receipts.schemas import JobStartedReceiptPayload, Receipt, ReceiptType
from compute_horde.utils import sign_blob

_USE_RANDOM = object()


def random_receipt(
    miner_keypair: Keypair = _USE_RANDOM,
    validator_keypair: Keypair = _USE_RANDOM,
):
    if miner_keypair is _USE_RANDOM:
        miner_keypair = random_keypair()

    if validator_keypair is _USE_RANDOM:
        validator_keypair = random_keypair()

    payload = JobStartedReceiptPayload(
        job_uuid=str(uuid4()),
        miner_hotkey=miner_keypair.ss58_address,
        validator_hotkey=validator_keypair.ss58_address,
        timestamp=timezone.now(),
        receipt_type=ReceiptType.JobStartedReceipt,
        executor_class=ExecutorClass.always_on__gpu_24gb,
        ttl=100,
        is_organic=False,
    )
    blob = payload.blob_for_signing()
    return Receipt(
        payload=payload,
        miner_signature=sign_blob(miner_keypair, blob),
        validator_signature=sign_blob(validator_keypair, blob),
    )


def random_keypair() -> Keypair:
    return Keypair.create_from_seed(random.randbytes(32).hex())



import abc


class TransportConnectionError(Exception):
    pass


class AbstractTransport(abc.ABC):
    """
    A base communication layer for sending and receiving messages.
    """

    def __init__(self, name: str, *args, **kwargs):
        self.name = name

    @abc.abstractmethod
    async def send(self, data: str | bytes) -> None:
        pass

    @abc.abstractmethod
    async def receive(self) -> str | bytes:
        pass

    @abc.abstractmethod
    async def start(self) -> None:
        pass

    @abc.abstractmethod
    async def stop(self) -> None:
        pass

    def __aiter__(self):
        return self

    async def __anext__(self):
        message = await self.receive()
        return message



import datetime

import bittensor
import pytest
import responses
from bittensor import Keypair

from compute_horde.executor_class import DEFAULT_EXECUTOR_CLASS
from compute_horde.receipts.schemas import (
    JobAcceptedReceiptPayload,
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
    Receipt,
)
from compute_horde.utils import sign_blob


@pytest.fixture
def keypair():
    return Keypair.create_from_mnemonic(
        "edit evoke caught tunnel harsh plug august group enact cable govern immense"
    )


@pytest.fixture
def validator_keypair():
    return Keypair.create_from_mnemonic(
        "slot excuse valid grief praise rifle spoil auction weasel glove pen share"
    )


@pytest.fixture
def miner_keypair():
    return Keypair.create_from_mnemonic(
        "almost fatigue race slim picnic mass better clog deal solve already champion"
    )


@pytest.fixture
def signature_wallet():
    wallet = bittensor.wallet(name="test_signature")
    # workaround the overwrite flag
    wallet.regenerate_coldkey(seed="8" * 64, use_password=False, overwrite=True)
    wallet.regenerate_hotkey(seed="9" * 64, use_password=False, overwrite=True)
    return wallet


@pytest.fixture
def receipts(validator_keypair, miner_keypair):
    payload1 = JobStartedReceiptPayload(
        job_uuid="3342460e-4a99-438b-8757-795f4cb348dd",
        miner_hotkey=miner_keypair.ss58_address,
        validator_hotkey=validator_keypair.ss58_address,
        timestamp=datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=datetime.UTC),
        executor_class=DEFAULT_EXECUTOR_CLASS,
        is_organic=False,
        ttl=30,
    )
    receipt1 = Receipt(
        payload=payload1,
        validator_signature=sign_blob(validator_keypair, payload1.blob_for_signing()),
        miner_signature=sign_blob(miner_keypair, payload1.blob_for_signing()),
    )

    payload2 = JobAcceptedReceiptPayload(
        job_uuid="3342460e-4a99-438b-8757-795f4cb348dd",
        miner_hotkey=miner_keypair.ss58_address,
        validator_hotkey=validator_keypair.ss58_address,
        timestamp=datetime.datetime(2020, 1, 1, 0, 5, 0, tzinfo=datetime.UTC),
        time_accepted=datetime.datetime(2020, 1, 1, 0, 4, 0, tzinfo=datetime.UTC),
        ttl=300,
    )
    receipt2 = Receipt(
        payload=payload2,
        validator_signature=sign_blob(validator_keypair, payload2.blob_for_signing()),
        miner_signature=sign_blob(miner_keypair, payload2.blob_for_signing()),
    )

    payload3 = JobFinishedReceiptPayload(
        job_uuid="3342460e-4a99-438b-8757-795f4cb348dd",
        miner_hotkey=miner_keypair.ss58_address,
        validator_hotkey=validator_keypair.ss58_address,
        timestamp=datetime.datetime(2020, 1, 1, 0, 10, 0, tzinfo=datetime.UTC),
        time_started=datetime.datetime(2020, 1, 1, 0, 9, 0, tzinfo=datetime.UTC),
        time_took_us=60_000_000,
        score_str="2.00",
    )
    receipt3 = Receipt(
        payload=payload3,
        validator_signature=sign_blob(validator_keypair, payload3.blob_for_signing()),
        miner_signature=sign_blob(miner_keypair, payload3.blob_for_signing()),
    )

    return [receipt1, receipt2, receipt3]


@pytest.fixture
def mocked_responses():
    with responses.RequestsMock() as rsps:
        yield rsps



class AddOnlyAdminMixin:
    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False


class ReadOnlyAdminMixin:
    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_add_permission(self, *args, **kwargs):
        return False



from .base import AbstractTransport, TransportConnectionError
from .stub import StubTransport
from .ws import WSTransport

__all__ = [
    "AbstractTransport",
    "TransportConnectionError",
    "WSTransport",
    "StubTransport",
]



import logging

from celery import shared_task
from django.apps import apps
from django.conf import settings

from compute_horde.utils import get_validators

logger = logging.getLogger(__name__)


@shared_task()
def fetch_validators():
    Validator = apps.get_model(settings.COMPUTE_HORDE_VALIDATOR_MODEL)
    key_field = settings.COMPUTE_HORDE_VALIDATOR_KEY_FIELD
    active_field = settings.COMPUTE_HORDE_VALIDATOR_ACTIVE_FIELD

    debug_validator_keys = set()
    if debug_field := getattr(settings, "COMPUTE_HORDE_VALIDATOR_DEBUG_FIELD", None):
        debug_validator_keys = set(
            Validator.objects.filter(**{debug_field: True, active_field: True}).values_list(
                key_field, flat=True
            )
        )

    validators = get_validators(
        netuid=settings.BITTENSOR_NETUID, network=settings.BITTENSOR_NETWORK
    )
    validator_keys = {v.hotkey for v in validators} | debug_validator_keys

    to_activate = []
    to_deactivate = []
    to_create = []
    for validator in Validator.objects.all():
        public_key = getattr(validator, key_field)
        if public_key in validator_keys:
            to_activate.append(validator)
            setattr(validator, active_field, True)
            validator_keys.remove(public_key)
        else:
            setattr(validator, active_field, False)
            to_deactivate.append(validator)

    for key in validator_keys:
        to_create.append(Validator(**{key_field: key, active_field: True}))

    Validator.objects.bulk_create(to_create)
    Validator.objects.bulk_update(to_activate + to_deactivate, [active_field])
    logger.info(
        f"Fetched validators. Activated: {len(to_activate)}, deactivated: {len(to_deactivate)}, "
        f"created: {len(to_create)}"
    )



import datetime
from datetime import timedelta
from typing import ClassVar, Self, TypeAlias, assert_never

from compute_horde_core.executor_class import ExecutorClass
from django.db import models

from compute_horde.executor_class import DEFAULT_EXECUTOR_CLASS
from compute_horde.receipts import ReceiptType
from compute_horde.receipts.schemas import (
    JobAcceptedReceiptPayload,
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
    Receipt,
)


class ReceiptNotSigned(Exception):
    pass


class AbstractReceipt(models.Model):
    job_uuid = models.UUIDField()
    validator_hotkey = models.CharField(max_length=256)
    miner_hotkey = models.CharField(max_length=256)
    validator_signature = models.CharField(max_length=256)
    miner_signature = models.CharField(max_length=256, null=True, blank=True)
    timestamp = models.DateTimeField()

    # https://github.com/typeddjango/django-stubs/issues/1684#issuecomment-1706446344
    objects: ClassVar[models.Manager[Self]]

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(fields=["job_uuid"], name="receipts_unique_%(class)s_job_uuid"),
        ]
        indexes = [
            models.Index(fields=["timestamp"], name="%(class)s_ts_idx"),
        ]

    def __str__(self):
        return f"job_uuid: {self.job_uuid}"


class JobStartedReceiptQuerySet(models.QuerySet["JobStartedReceipt"]):
    def valid_at(self, dt: datetime.datetime):
        return self.annotate(
            valid_until=models.ExpressionWrapper(
                models.F("timestamp") + models.F("ttl") * timedelta(seconds=1),
                output_field=models.DateTimeField(),
            ),
        ).filter(
            timestamp__lte=dt,
            valid_until__gte=dt,
        )


class JobStartedReceipt(AbstractReceipt):
    executor_class = models.CharField(max_length=255, default=DEFAULT_EXECUTOR_CLASS)
    is_organic = models.BooleanField()
    ttl = models.IntegerField()

    objects = JobStartedReceiptQuerySet.as_manager()  # type: ignore

    def to_receipt(self) -> Receipt:
        if self.miner_signature is None:
            raise ReceiptNotSigned("Miner signature is required")

        return Receipt(
            payload=JobStartedReceiptPayload(
                job_uuid=str(self.job_uuid),
                miner_hotkey=self.miner_hotkey,
                validator_hotkey=self.validator_hotkey,
                timestamp=self.timestamp,
                executor_class=ExecutorClass(self.executor_class),
                is_organic=self.is_organic,
                ttl=self.ttl,
            ),
            validator_signature=self.validator_signature,
            miner_signature=self.miner_signature,
        )

    @classmethod
    def from_receipt(cls, receipt: Receipt) -> "JobStartedReceipt":
        if not isinstance(receipt.payload, JobStartedReceiptPayload):
            raise ValueError(
                f"Incompatible receipt payload type. "
                f"Got: {type(receipt.payload).__name__} "
                f"Expected: {JobStartedReceiptPayload.__name__}"
            )

        return cls.from_payload(
            receipt.payload, receipt.validator_signature, receipt.miner_signature
        )

    @classmethod
    def from_payload(
        cls,
        payload: JobStartedReceiptPayload,
        validator_signature: str,
        miner_signature: str | None = None,
    ) -> "JobStartedReceipt":
        return JobStartedReceipt(
            job_uuid=payload.job_uuid,
            miner_hotkey=payload.miner_hotkey,
            validator_signature=validator_signature,
            miner_signature=miner_signature,
            timestamp=payload.timestamp,
            executor_class=payload.executor_class,
            is_organic=payload.is_organic,
            ttl=payload.ttl,
        )


class JobAcceptedReceipt(AbstractReceipt):
    time_accepted = models.DateTimeField()
    ttl = models.IntegerField()

    def to_receipt(self) -> Receipt:
        if self.miner_signature is None:
            raise ReceiptNotSigned("Miner signature is required")

        return Receipt(
            payload=JobAcceptedReceiptPayload(
                job_uuid=str(self.job_uuid),
                miner_hotkey=self.miner_hotkey,
                validator_hotkey=self.validator_hotkey,
                timestamp=self.timestamp,
                time_accepted=self.time_accepted,
                ttl=self.ttl,
            ),
            validator_signature=self.validator_signature,
            miner_signature=self.miner_signature,
        )

    @classmethod
    def from_receipt(cls, receipt: Receipt) -> "JobAcceptedReceipt":
        if not isinstance(receipt.payload, JobAcceptedReceiptPayload):
            raise ValueError(
                f"Incompatible receipt payload type. "
                f"Got: {type(receipt.payload).__name__} "
                f"Expected: {JobAcceptedReceiptPayload.__name__}"
            )

        return cls.from_payload(
            receipt.payload, receipt.validator_signature, receipt.miner_signature
        )

    @classmethod
    def from_payload(
        cls,
        payload: JobAcceptedReceiptPayload,
        validator_signature: str,
        miner_signature: str | None = None,
    ) -> "JobAcceptedReceipt":
        return JobAcceptedReceipt(
            job_uuid=payload.job_uuid,
            miner_hotkey=payload.miner_hotkey,
            validator_signature=validator_signature,
            miner_signature=miner_signature,
            timestamp=payload.timestamp,
            time_accepted=payload.time_accepted,
            ttl=payload.ttl,
        )


class JobFinishedReceipt(AbstractReceipt):
    time_started = models.DateTimeField()
    time_took_us = models.BigIntegerField()
    score_str = models.CharField(max_length=256)

    def time_took(self):
        return timedelta(microseconds=self.time_took_us)

    def score(self):
        return float(self.score_str)

    def to_receipt(self) -> Receipt:
        if self.miner_signature is None:
            raise ReceiptNotSigned("Miner signature is required")

        return Receipt(
            payload=JobFinishedReceiptPayload(
                job_uuid=str(self.job_uuid),
                miner_hotkey=self.miner_hotkey,
                validator_hotkey=self.validator_hotkey,
                timestamp=self.timestamp,
                time_started=self.time_started,
                time_took_us=self.time_took_us,
                score_str=self.score_str,
            ),
            validator_signature=self.validator_signature,
            miner_signature=self.miner_signature,
        )

    @classmethod
    def from_receipt(cls, receipt: Receipt) -> "JobFinishedReceipt":
        if not isinstance(receipt.payload, JobFinishedReceiptPayload):
            raise ValueError(
                f"Incompatible receipt payload type. "
                f"Got: {receipt.payload.__class__.__name__} "
                f"Expected: {JobFinishedReceiptPayload.__name__}"
            )

        return cls.from_payload(
            receipt.payload, receipt.validator_signature, receipt.miner_signature
        )

    @classmethod
    def from_payload(
        cls,
        payload: JobFinishedReceiptPayload,
        validator_signature: str,
        miner_signature: str | None = None,
    ) -> "JobFinishedReceipt":
        return JobFinishedReceipt(
            job_uuid=payload.job_uuid,
            miner_hotkey=payload.miner_hotkey,
            validator_signature=validator_signature,
            miner_signature=miner_signature,
            timestamp=payload.timestamp,
            time_started=payload.time_started,
            time_took_us=payload.time_took_us,
            score_str=payload.score_str,
        )


ReceiptModel: TypeAlias = JobAcceptedReceipt | JobStartedReceipt | JobFinishedReceipt


def receipt_to_django_model(receipt: Receipt) -> ReceiptModel:
    match receipt.payload.receipt_type:
        case ReceiptType.JobAcceptedReceipt:
            return JobAcceptedReceipt.from_receipt(receipt)
        case ReceiptType.JobStartedReceipt:
            return JobStartedReceipt.from_receipt(receipt)
        case ReceiptType.JobFinishedReceipt:
            return JobFinishedReceipt.from_receipt(receipt)
        case _:
            assert_never(receipt.payload.receipt_type)



import pytest
from compute_horde_core.executor_class import ExecutorClass
from compute_horde_core.streaming import StreamingDetails
from pydantic import TypeAdapter

from compute_horde.miner_client.organic import (
    OrganicJobDetails,
    OrganicMinerClient,
    execute_organic_job_on_miner,
)
from compute_horde.protocol_messages import (
    V0AcceptJobRequest,
    V0ExecutionDoneRequest,
    V0ExecutorReadyRequest,
    V0InitialJobRequest,
    V0JobAcceptedReceiptRequest,
    V0JobFinishedReceiptRequest,
    V0JobFinishedRequest,
    V0JobRequest,
    V0StreamingJobReadyRequest,
    V0VolumesReadyRequest,
    ValidatorAuthForMiner,
    ValidatorToMinerMessage,
)
from compute_horde.transport import StubTransport

JOB_UUID = "b4793a02-33a2-4a49-b4e2-4a7b903847e7"


class MinerStubTransport(StubTransport):
    def __init__(self, name: str, messages: list[str], *args, **kwargs):
        super().__init__(name, messages, *args, **kwargs)
        self.sent_models = []

    async def send(self, message):
        await super().send(message)
        self.sent_models.append(TypeAdapter(ValidatorToMinerMessage).validate_json(message))


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_run_organic_job__success(keypair):
    mock_transport = MinerStubTransport(
        "mock",
        [
            V0AcceptJobRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0ExecutorReadyRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0VolumesReadyRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0ExecutionDoneRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0JobFinishedRequest(
                job_uuid=JOB_UUID,
                docker_process_stdout="stdout",
                docker_process_stderr="stderr",
                artifacts={},
            ).model_dump_json(),
        ],
    )
    client = OrganicMinerClient(
        miner_hotkey="mock",
        miner_address="0.0.0.0",
        miner_port=1234,
        job_uuid=JOB_UUID,
        my_keypair=keypair,
        transport=mock_transport,
    )
    job_details = OrganicJobDetails(
        job_uuid=JOB_UUID,
        executor_class=ExecutorClass.always_on__llm__a6000,
        docker_image="mock",
    )
    stdout, stderr, artifacts, upload_results = await execute_organic_job_on_miner(
        client, job_details, reservation_time_limit=2, executor_startup_time_limit=2
    )

    assert stdout == "stdout"
    assert stderr == "stderr"

    sent_models_types = [type(model) for model in mock_transport.sent_models]
    assert sent_models_types == [
        ValidatorAuthForMiner,
        V0InitialJobRequest,
        V0JobAcceptedReceiptRequest,
        V0JobRequest,
        V0JobFinishedReceiptRequest,
    ]


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_run_organic_job_streaming__success(keypair):
    mock_transport = MinerStubTransport(
        "mock",
        [
            V0AcceptJobRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0ExecutorReadyRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0VolumesReadyRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0StreamingJobReadyRequest(
                job_uuid=JOB_UUID,
                public_key="dummy-cert",
                port=12345,
            ).model_dump_json(),
            V0ExecutionDoneRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0JobFinishedRequest(
                job_uuid=JOB_UUID,
                docker_process_stdout="streaming-stdout",
                docker_process_stderr="streaming-stderr",
                artifacts={},
            ).model_dump_json(),
        ],
    )
    client = OrganicMinerClient(
        miner_hotkey="mock",
        miner_address="0.0.0.0",
        miner_port=1234,
        job_uuid=JOB_UUID,
        my_keypair=keypair,
        transport=mock_transport,
    )
    job_details = OrganicJobDetails(
        job_uuid=JOB_UUID,
        executor_class=ExecutorClass.always_on__llm__a6000,
        docker_image="mock",
        streaming_details=StreamingDetails(public_key="dummy-cert"),
    )
    await execute_organic_job_on_miner(
        client, job_details, reservation_time_limit=2, executor_startup_time_limit=2
    )

    # Find the initial job request sent by the client
    initial_job_request = next(
        (m for m in mock_transport.sent_models if isinstance(m, V0InitialJobRequest)), None
    )
    assert initial_job_request is not None, "Initial job request was not sent"
    assert initial_job_request.streaming_details is not None, (
        "Streaming details not sent in initial job request"
    )
    assert initial_job_request.streaming_details.public_key == "dummy-cert"


# TODO:
#   - unhappy path
#       - connection error
#       - initial_response timeout
#       - send error event callback
#       - declined
#       - executor failed
#       - full_response timeout
#       - job failed



import pytest
from freezegun import freeze_time

from compute_horde.utils import Timer


def test_timer_none():
    with freeze_time("2024-01-01 00:00:00"):
        timer = Timer()
        with pytest.raises(ValueError):
            timer.time_left()


def test_timer_timeout():
    with freeze_time("2024-01-01 00:00:00") as frozen_time:
        timer = Timer(timeout=300)
        frozen_time.move_to("2024-01-01 00:00:30")
        assert timer.passed_time() == 30.0
        assert timer.time_left() == 270.0



import gzip
import logging
import os
import re
import shutil
import tempfile
from collections import defaultdict
from collections.abc import Sequence
from datetime import datetime
from glob import glob
from pathlib import Path

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone

from compute_horde.receipts.schemas import (
    Receipt,
)
from compute_horde.receipts.store.base import BaseReceiptStore

"""
Considerations:
- smaller pages make for more (smaller) batches when downloading them for the first time (and we need ~5 hours of pages)
- pages get copied over on write, larger pages may impact efficiency of this
- we keep up with the 2 latest pages based on time and the clock may be out of sync between miners and validators
"""
PAGE_TIME = 60 * 5  # 5 minutes
N_ACTIVE_PAGES = 2

logger = logging.getLogger(__name__)


class LocalFilesystemPagedReceiptStore(BaseReceiptStore):
    def __init__(self):
        super().__init__()
        if not getattr(settings, "LOCAL_RECEIPTS_ROOT", ""):
            raise ImproperlyConfigured("Required settings.py setting missing: LOCAL_RECEIPTS_ROOT")
        self.pages_directory = Path(settings.LOCAL_RECEIPTS_ROOT)  # type: ignore
        self.pages_directory.mkdir(parents=True, exist_ok=True)

    @classmethod
    def current_page(cls) -> int:
        """
        Get current page ID
        """
        return cls.current_page_at(timezone.now())

    @staticmethod
    def current_page_at(dt: datetime) -> int:
        """
        Calculate what the current page was at given time
        """
        return int(dt.timestamp() // PAGE_TIME)

    def store(self, receipts: Sequence[Receipt]) -> None:
        """
        Append receipts to the store.
        """
        pages: defaultdict[int, list[Receipt]] = defaultdict(list)
        for receipt in receipts:
            pages[self.current_page()].append(receipt)
        for page, receipts_in_page in pages.items():
            self._append_to_page(receipts_in_page, page)

    def page_filepath(self, page: int) -> Path:
        """
        Find the filepath under which given pagefile should be found.
        Does not check whether it actually exists or not.
        """
        return self.pages_directory / f"{page}.jsonl"

    def archive_filepath(self, page: int) -> Path:
        """
        Find the filepath under which given pagefile's archive should be found.
        Does not check whether it actually exists or not.
        """
        return self.pages_directory / f"{page}.jsonl.gz"

    def delete_page(self, page: int) -> None:
        """
        Deletes the page file from the file system if it exists.
        Does nothing otherwise.
        """
        try:
            os.unlink(self.page_filepath(page))
        except FileNotFoundError:
            pass
        try:
            os.unlink(self.archive_filepath(page))
        except FileNotFoundError:
            pass

    def evict(self, cutoff: datetime) -> None:
        self.delete_pages_older_than(cutoff)

    def delete_pages_older_than(self, older_than: int | datetime) -> None:
        if isinstance(older_than, datetime):
            older_than = self.current_page_at(older_than)

        old_pages = [p for p in self.get_available_pages() if p < older_than]

        for old_page in old_pages:
            try:
                self.delete_page(old_page)
            except Exception:
                logger.exception("Error while deleting page %s", old_page, exc_info=True)

    def archive_old_pages(self) -> None:
        """
        Create archives for all old pages if they don't exist yet.
        Skips active pages as these can be still written to.
        """
        current_page = self.current_page()
        upper_cutoff = current_page - N_ACTIVE_PAGES
        pages_to_archive = [p for p in self.get_available_pages() if p <= upper_cutoff]
        for page in pages_to_archive:
            archive_filepath = self.archive_filepath(page)
            if archive_filepath.exists():
                continue
            self.do_archive_page(page)

    def do_archive_page(self, page: int) -> None:
        """
        Packs given page and creates an additional archive file used by nginx to serve the page.
        """
        with open(self.page_filepath(page), "rb") as page_file:
            with gzip.open(self.archive_filepath(page), "wb") as archive_file:
                shutil.copyfileobj(page_file, archive_file)

    def get_available_pages(self) -> list[int]:
        """
        Return IDs of all existing pages.
        """
        pagefiles = self._get_available_page_filepaths()
        pages: list[int] = []
        pattern = re.compile(r"(\d+)\.jsonl$")
        for pagefile in pagefiles:
            match = pattern.search(str(pagefile))
            if match:
                pages.append(int(match[1]))
        return pages

    def _append_to_page(self, receipts: Sequence[Receipt], page: int) -> None:
        """
        Write new receipts to the specified page.
        For read safety this will copy the page file first, append to the copy and do a swap with the original.
        """
        page_filepath = self.page_filepath(page)
        page_filepath.touch(exist_ok=True)
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_page_filepath = Path(tmpdir) / "tmp_page.jsonl"
            shutil.copyfile(page_filepath, tmp_page_filepath)
            with open(tmp_page_filepath, "a") as pagefile:
                for r in receipts:
                    pagefile.write(r.model_dump_json())
                    pagefile.write("\n")
            shutil.move(tmp_page_filepath, page_filepath)

    def _get_available_page_filepaths(self) -> list[Path]:
        """
        Return filepaths of all existing pages.
        """
        pagefiles = glob(str(self.pages_directory / "*.jsonl"))
        return [Path(pagefile) for pagefile in pagefiles]



from argparse import Namespace

import pytest
from freezegun import freeze_time

from compute_horde.dynamic_config import sync_dynamic_config


def test_dynamic_config__non_dynamic_key_skipped(mocked_responses):
    # arrange
    config_url = "http://127.0.0.1:8000/config.json"
    mocked_responses.get(
        config_url,
        json={
            "RANDOM_KEY": {
                "description": "...",
                "items": [{"value": 1}],
            },
            "DYNAMIC_KEY": {
                "description": "...",
                "items": [{"value": 2}],
            },
        },
    )
    namespace = Namespace()

    # act
    sync_dynamic_config(config_url, namespace)

    # assert
    assert vars(namespace) == {"DYNAMIC_KEY": 2}


@pytest.mark.parametrize(
    ("frozen_time", "expected_value"),
    (
        ("2024-01-02T00:00:00.000Z", 2),
        ("2024-01-03T00:01:01.000Z", 3),
        ("2024-01-04T01:01:01.000Z", 4),
    ),
)
def test_dynamic_config__correct_time_is_picked(mocked_responses, frozen_time, expected_value):
    # arrange
    config_url = "http://127.0.0.1:8000/config.json"
    mocked_responses.get(
        config_url,
        json={
            "DYNAMIC_KEY": {
                "description": "...",
                "items": [
                    {
                        "value": i,
                        "effective_from": f"2024-01-0{i}T00:00:00.000Z",
                    }
                    for i in range(1, 6)
                ],
            }
        },
    )
    namespace = Namespace()

    # act
    with freeze_time(frozen_time):
        sync_dynamic_config(config_url, namespace)

    # assert
    assert vars(namespace) == {"DYNAMIC_KEY": expected_value}


def test_dynamic_config__json_value(mocked_responses):
    # arrange
    config_url = "http://127.0.0.1:8000/config.json"
    mocked_responses.get(
        config_url,
        json={
            "DYNAMIC_KEY": {
                "description": "...",
                "items": [{"value": {"key": [1, 2, 3]}}],
            },
        },
    )
    namespace = Namespace()

    # act
    sync_dynamic_config(config_url, namespace)

    # assert
    assert vars(namespace) == {"DYNAMIC_KEY": {"key": [1, 2, 3]}}



from django.db import models


class Validator(models.Model):
    public_key = models.TextField(unique=True)
    active = models.BooleanField()
    debug = models.BooleanField(default=False)

    def __str__(self):
        return f"hotkey: {self.public_key}"


# stats-collector uses this definition of the validator model
class Validator2(models.Model):
    ss58_address = models.CharField(max_length=48, unique=True)
    is_active = models.BooleanField()

    def __str__(self):
        return f"hotkey: {self.ss58_address}"



import base64
import uuid

from compute_horde_core.signature import BittensorWalletSigner, BittensorWalletVerifier, Signature
from compute_horde_core.volume import VolumeType, ZipUrlVolume

from compute_horde.executor_class import DEFAULT_EXECUTOR_CLASS
from compute_horde.fv_protocol.facilitator_requests import V2JobRequest


def test_signed_job_roundtrip(signature_wallet):
    volume = ZipUrlVolume(
        volume_type=VolumeType.zip_url,
        contents="https://example.com/input.zip",
        relative_path="input",
    )
    job = V2JobRequest(
        uuid=str(uuid.uuid4()),
        executor_class=DEFAULT_EXECUTOR_CLASS,
        docker_image="hello-world",
        args=["--verbose", "--dry-run"],
        env={"CUDA": "1"},
        use_gpu=False,
        volume=volume,
        output_upload=None,
        download_time_limit=1,
        execution_time_limit=1,
        upload_time_limit=1,
        streaming_start_time_limit=1,
    )

    signer = BittensorWalletSigner(signature_wallet)
    payload = job.json_for_signing()
    raw_signature = signer.sign(payload)

    job.signature = Signature(
        signature_type=raw_signature.signature_type,
        signatory=raw_signature.signatory,
        timestamp_ns=raw_signature.timestamp_ns,
        signature=base64.b64encode(raw_signature.signature),
    )

    job_json = job.model_dump_json()
    deserialized_job = V2JobRequest.model_validate_json(job_json)

    assert deserialized_job.signature is not None
    deserialized_raw_signature = Signature(
        signature_type=deserialized_job.signature.signature_type,
        signatory=deserialized_job.signature.signatory,
        timestamp_ns=deserialized_job.signature.timestamp_ns,
        signature=base64.b64encode(deserialized_job.signature.signature),
    )

    deserialized_payload = deserialized_job.json_for_signing()
    verifier = BittensorWalletVerifier()
    verifier.verify(deserialized_payload, deserialized_raw_signature)



import json
import typing
from typing import Annotated, Literal, TypeAlias

import pydantic
from compute_horde_core.executor_class import ExecutorClass
from compute_horde_core.output_upload import MultiUpload, OutputUpload
from compute_horde_core.signature import SignedFields, SignedRequest
from compute_horde_core.streaming import StreamingDetails
from compute_horde_core.volume import MultiVolume, Volume
from pydantic import BaseModel, JsonValue


class Error(BaseModel, extra="allow"):
    msg: str
    type: str
    help: str = ""


class Response(BaseModel, extra="forbid"):
    """Message sent from facilitator to validator in response to AuthenticationRequest & JobStatusUpdate"""

    status: Literal["error", "success"]
    errors: list[Error] = []


class V0JobCheated(SignedRequest, BaseModel, extra="forbid"):
    """Message sent from facilitator to report cheated job"""

    # this points to a `ValidatorConsumer.job_cheated` handler (fuck you django-channels!)
    type: Literal["job.cheated"] = "job.cheated"
    message_type: Literal["V0JobCheated"] = "V0JobCheated"

    job_uuid: str

    def get_signed_payload(self) -> JsonValue:
        return json.dumps({"job_uuid": self.job_uuid})


def to_json_array(data) -> list[JsonValue]:
    return typing.cast(list[JsonValue], [x.model_dump() for x in data])


class V2JobRequest(SignedRequest, BaseModel, extra="forbid"):
    """Message sent from facilitator to validator to request a job execution"""

    # this points to a `ValidatorConsumer.job_new` handler (fuck you django-channels!)
    type: Literal["job.new"] = "job.new"
    message_type: Literal["V2JobRequest"] = "V2JobRequest"

    uuid: str

    # !!! all fields below are included in the signed json payload
    executor_class: ExecutorClass
    docker_image: str
    args: list[str]
    env: dict[str, str]
    use_gpu: bool
    volume: Volume | None = None
    output_upload: OutputUpload | None = None
    artifacts_dir: str | None = None
    on_trusted_miner: bool = False
    download_time_limit: int
    execution_time_limit: int
    upload_time_limit: int
    streaming_start_time_limit: int
    streaming_details: StreamingDetails | None = None
    # !!! all fields above are included in the signed json payload

    def get_args(self):
        return self.args

    def get_signed_fields(self) -> SignedFields:
        volumes = (
            to_json_array(
                self.volume.volumes if isinstance(self.volume, MultiVolume) else [self.volume]
            )
            if self.volume
            else []
        )

        uploads = (
            to_json_array(
                self.output_upload.uploads
                if isinstance(self.output_upload, MultiUpload)
                # TODO: fix consolidate faci output_upload types
                else [self.output_upload]
            )
            if self.output_upload
            else []
        )

        signed_fields = SignedFields(
            executor_class=self.executor_class,
            docker_image=self.docker_image,
            args=self.args,
            env=self.env,
            use_gpu=self.use_gpu,
            artifacts_dir=self.artifacts_dir or "",
            on_trusted_miner=self.on_trusted_miner,
            volumes=volumes,
            uploads=uploads,
            download_time_limit=self.download_time_limit,
            execution_time_limit=self.execution_time_limit,
            upload_time_limit=self.upload_time_limit,
            streaming_start_time_limit=self.streaming_start_time_limit,
            streaming_details=self.streaming_details,
        )
        return signed_fields

    def get_signed_payload(self) -> JsonValue:
        return self.get_signed_fields().model_dump_json()  # type: ignore

    def json_for_signing(self) -> JsonValue:
        payload = self.model_dump(mode="json")
        del payload["type"]
        del payload["message_type"]
        del payload["signature"]
        return payload


OrganicJobRequest: TypeAlias = Annotated[
    V2JobRequest,
    pydantic.Field(discriminator="message_type"),
]



import asyncio
import os
import subprocess
import sys
import uuid

import pytest
import requests
from compute_horde_core.certificate import (
    generate_certificate,
    generate_certificate_at,
    save_public_key,
    start_nginx,
    write_certificate,
)
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

NGINX_PORT = 8443
NGINX_URI = f"https://localhost:{NGINX_PORT}"

NGINX_CONF = """
http {
    server {
        listen 443 ssl;
        server_name example.com;

        ssl_certificate /etc/nginx/ssl/certificate.pem;
        ssl_certificate_key /etc/nginx/ssl/private_key.pem;

        ssl_client_certificate /etc/nginx/ssl/client.crt;
        ssl_verify_client on;

        location / {
            if ($ssl_client_verify != SUCCESS) { return 403; }
            return 200 'Hello World!';
            add_header Content-Type text/plain;
        }
    }
    
    server {
        listen 80;
        server_name localhost;

        # for checking if nginx is running
        location /ok { return 200; }
    }
}

events {
    worker_connections 1024;
}
"""


def test_generate_ca_and_server_certs(tmp_path):
    generate_certificate_at(tmp_path, "blabla")
    tmp_path = tmp_path / "ssl"

    assert set(os.listdir(tmp_path)) == {"certificate.pem", "private_key.pem"}

    # check the cert is valid
    x509.load_pem_x509_certificate((tmp_path / "certificate.pem").read_bytes())

    # check the public key is valid
    key = serialization.load_pem_private_key((tmp_path / "private_key.pem").read_bytes(), None)
    assert isinstance(key, RSAPrivateKey)


async def start_nginx_with_certificates(
    nginx_conf: str,
    public_key: str,
    port: int,
    container_name: str,
) -> tuple[str, str]:
    tmp_path, _, cert = generate_certificate_at()
    save_public_key(public_key, tmp_path)

    job_network = f"{container_name}_network"
    process = await asyncio.create_subprocess_exec(
        "docker", "network", "create", "--internal", job_network
    )
    await process.wait()

    await start_nginx(
        nginx_conf, port, tmp_path, job_network=job_network, container_name=container_name
    )
    return cert


@pytest.fixture
def container_name():
    return str(uuid.uuid4())


@pytest.fixture
def cleanup_docker(container_name):
    yield
    subprocess.run(["docker", "kill", container_name])
    subprocess.run(["docker", "network", "rm", f"{container_name}_network"])


@pytest.mark.asyncio
@pytest.mark.skipif(sys.platform != "linux", reason="requires linux")
async def test_certificates_with_nginx__success(tmp_path, container_name, cleanup_docker):
    _, public_key, cert = generate_certificate_at(tmp_path)
    nginx_cert_path, _ = await start_nginx_with_certificates(
        NGINX_CONF, public_key, NGINX_PORT, container_name
    )

    resp = requests.get(NGINX_URI, verify=nginx_cert_path, cert=cert)
    assert resp.status_code == 200
    assert resp.text.strip() == "Hello World!"


@pytest.mark.asyncio
@pytest.mark.skipif(sys.platform != "linux", reason="requires linux")
async def test_certificates_with_nginx__no_cert(tmp_path, container_name, cleanup_docker):
    _, public_key, _ = generate_certificate_at(tmp_path)
    nginx_cert_path, _ = await start_nginx_with_certificates(
        NGINX_CONF, public_key, NGINX_PORT, container_name
    )

    resp = requests.get(NGINX_URI, verify=nginx_cert_path)

    # 400 no cert, 403 wrong cert
    assert resp.status_code != 200


@pytest.mark.asyncio
@pytest.mark.skipif(sys.platform != "linux", reason="requires linux")
async def test_certificates_with_nginx__fail(tmp_path, container_name, cleanup_docker):
    _, public_key, _ = generate_certificate_at(tmp_path)
    _ = await start_nginx_with_certificates(NGINX_CONF, public_key, NGINX_PORT, container_name)

    cert_path = tmp_path / "wrong_certificate.pem"
    certificate, _ = generate_certificate("127.0.0.1")
    write_certificate(certificate, cert_path)

    with pytest.raises(requests.exceptions.SSLError):
        requests.get(NGINX_URI, verify=str(cert_path))



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest
from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings
from freezegun import freeze_time

from compute_horde.receipts.store.local import (
    N_ACTIVE_PAGES,
    PAGE_TIME,
    LocalFilesystemPagedReceiptStore,
)
from tests.utils import random_receipt


@pytest.mark.parametrize("bad_value", [None, ""])
def test_exits_when_settings_missing(bad_value):
    with override_settings(LOCAL_RECEIPTS_ROOT=bad_value):
        with pytest.raises(ImproperlyConfigured):
            LocalFilesystemPagedReceiptStore().store([random_receipt()])


def test_creates_directory():
    with TemporaryDirectory() as tmpdir:
        subdir = Path(tmpdir) / "subdir"
        with override_settings(LOCAL_RECEIPTS_ROOT=str(subdir)):
            store = LocalFilesystemPagedReceiptStore()
            store.store([random_receipt()])

        assert subdir.exists()


def test_page_advances_with_time(time_freezer):
    initial_page = LocalFilesystemPagedReceiptStore.current_page()
    time_freezer.tick(delta=PAGE_TIME)
    next_page = LocalFilesystemPagedReceiptStore.current_page()
    assert next_page == initial_page + 1


def test_receipts_are_written_to_current_page(local_store, receipts_directory):
    with patch.object(LocalFilesystemPagedReceiptStore, "current_page") as current_page:
        for i in range(1, 6):
            current_page.return_value = i
            local_store.store([random_receipt() for _ in range(10)])

    for i in range(1, 6):
        assert (receipts_directory / f"{i}.jsonl").exists()


def test_receipts_are_readable(local_store):
    current_page = LocalFilesystemPagedReceiptStore.current_page()
    current_page_filepath = local_store.page_filepath(current_page)
    expected_receipts = []
    for _ in range(5):
        batch_to_write = [random_receipt() for _ in range(5)]
        local_store.store(batch_to_write)
        expected_receipts.extend(batch_to_write)
        with open(current_page_filepath, "rb") as pagefile:
            for file_line, receipt in zip(pagefile, expected_receipts):
                read_receipt = type(receipt).model_validate_json(file_line)
                assert read_receipt == receipt


def test_archives_correct_pages(local_store):
    with patch.object(LocalFilesystemPagedReceiptStore, "current_page") as current_page:
        for i in range(1, 10):
            current_page.return_value = i
            local_store.store([random_receipt()])

        local_store.archive_old_pages()

    for i in range(1, 10 - N_ACTIVE_PAGES):
        archive_filepath = local_store.archive_filepath(i)
        assert archive_filepath.exists()

    for i in range(10 - N_ACTIVE_PAGES + 1, 10):
        archive_filepath = local_store.archive_filepath(i)
        assert not archive_filepath.exists()


def test_lists_available_pages(local_store):
    with patch.object(LocalFilesystemPagedReceiptStore, "current_page") as current_page:
        expected_pages = [*range(1, 123, 7)]
        for page in expected_pages:
            current_page.return_value = page
            local_store.store([random_receipt()])

        assert sorted(local_store.get_available_pages()) == expected_pages


def test_deletes_old_pages(local_store):
    with patch.object(LocalFilesystemPagedReceiptStore, "current_page") as current_page:
        for page_to_create in range(1, 10):  # 1-9 inclusive
            current_page.return_value = page_to_create
            local_store.store([random_receipt()])

        assert sorted(local_store.get_available_pages()) == [*range(1, 10)]

        local_store.delete_pages_older_than(5)
        assert sorted(local_store.get_available_pages()) == [*range(5, 10)]


@pytest.fixture(autouse=True)
def time_freezer():
    with freeze_time("2024-12-01T00:00:00Z") as freezer:
        yield freezer


@pytest.fixture()
def receipts_directory():
    with TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)


@pytest.fixture()
def local_store(receipts_directory: Path):
    with override_settings(LOCAL_RECEIPTS_ROOT=str(receipts_directory)):
        yield LocalFilesystemPagedReceiptStore()



import logging
from abc import ABC, abstractmethod
from functools import cache

from django.conf import settings
from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger(__name__)


class TransferCheckpointBackend(ABC):
    @abstractmethod
    async def get(self, key: str) -> int: ...

    @abstractmethod
    async def set(self, key: str, checkpoint: int) -> None: ...


class DjangoCacheTransferCheckpointBackend(TransferCheckpointBackend):
    def __init__(self):
        if not getattr(settings, "RECEIPT_TRANSFER_CHECKPOINT_CACHE", ""):
            raise ImproperlyConfigured(
                "Required settings.py setting missing: RECEIPT_TRANSFER_CHECKPOINT_CACHE"
            )

        self.cache = caches[settings.RECEIPT_TRANSFER_CHECKPOINT_CACHE]  # type: ignore

    async def get(self, key: str) -> int:
        try:
            return int(await self.cache.aget(key, 0))
        except TypeError:
            logger.error(
                f"Django cache contained non-integer checkpoint value for {key}. Defaulting to 0."
            )
            return 0

    async def set(self, key: str, checkpoint: int) -> None:
        await self.cache.aset(key, str(checkpoint))


@cache
def checkpoint_backend() -> TransferCheckpointBackend:
    return DjangoCacheTransferCheckpointBackend()



import base64
import datetime

import freezegun
import pytest
from compute_horde_core.executor_class import ExecutorClass
from compute_horde_core.output_upload import SingleFilePutUpload
from compute_horde_core.signature import (
    BittensorWalletSigner,
    BittensorWalletVerifier,
    Signature,
    SignatureInvalidException,
    SignatureNotFound,
    SignedFields,
    hash_message_signature,
    signature_from_headers,
    signature_payload,
)
from compute_horde_core.volume import HuggingfaceVolume, MultiVolume, SingleFileVolume

from compute_horde.fv_protocol.facilitator_requests import (
    V2JobRequest,
    to_json_array,
)
from compute_horde.fv_protocol.validator_requests import V0AuthenticationRequest


@pytest.fixture
def sample_data():
    return {
        "b": 2,
        "array": [1, 2, 3],
        "dict": {"subdict:": {"null": None}},
    }


@pytest.fixture
def sample_signature():
    return Signature(
        signature_type="bittensor",
        signatory="5FUJCuGtQPonu8B9JKH4BwsKzEdtyBTpyvbBk2beNZ4iX8sk",  # hotkey address
        timestamp_ns=1718845323456788992,
        signature=base64.b64encode(
            base64.b85decode(
                "1SaAcLt*GG`2RG*@xmapXZ14m*Y`@b1MP(hAfEnwXkO5Os<30drw{`X`15JFP4GWR96T7p>rUmYA#=8Z"
            )
        ),
    )


def test_hash_message_signature(sample_data, sample_signature):
    assert hash_message_signature(sample_data, sample_signature) == base64.b85decode(
        "Dk7mtj^WM^#_n_!-C@$EHF*O@>;mdK%XS?515>IhxRvgxdT<nq$ImMmQ)jr-hpUSs#o+GQRX~t<q0-~1"
    )


@freezegun.freeze_time("2024-06-20T01:02:03.456789")
def test_bittensor_wallet_signer_sign(signature_wallet, sample_data):
    signer = BittensorWalletSigner(signature_wallet)
    signature = signer.sign(sample_data)

    assert signature == Signature(
        signature_type="bittensor",
        signatory=signature_wallet.hotkey.ss58_address,
        timestamp_ns=1718845323456788992,
        signature=base64.b64encode(signature.signature),
    )

    assert isinstance(signature.signature, bytes) and len(signature.signature) == 64


def test_bittensor_wallet_verifier_verify(sample_data, sample_signature):
    verifier = BittensorWalletVerifier()
    verifier.verify(sample_data, sample_signature)


def test_bittensor_wallet_verifier_verify_invalid_signature(sample_data, sample_signature):
    verifier = BittensorWalletVerifier()
    sample_signature.signature = b"invalid"
    with pytest.raises(SignatureInvalidException):
        verifier.verify(sample_data, sample_signature)


def test_bittensor_wallet_verifier_verify_signature_timeout(sample_data, sample_signature):
    verifier = BittensorWalletVerifier()

    with pytest.raises(SignatureInvalidException):
        verifier.verify(sample_data, sample_signature, newer_than=datetime.datetime.now())


@pytest.mark.parametrize(
    "data",
    [
        {"b": 2, "array": [1, 2, 3], "dict": {"subdict:": {"null": None}}},
        b"test",
    ],
)
def test_bittensor_wallet__sign_n_verify(signature_wallet, data):
    signer = BittensorWalletSigner(signature_wallet)
    verifier = BittensorWalletVerifier()

    signature = signer.sign(data)
    verifier.verify(data, signature)


def test_signature_from_headers__not_found():
    with pytest.raises(SignatureNotFound):
        signature_from_headers({})


def test_signature_payload():
    assert signature_payload(
        "get", "https://example.com/car", headers={"Date": "X"}, json={"a": 1}
    ) == {
        "action": "GET /car",
        "json": {"a": 1},
    }


def test_signed_fields__missing_fields():
    facilitator_request_json = {
        "executor_class": str(ExecutorClass.always_on__llm__a6000),
        "env": {},
        "use_gpu": False,
        "input_url": "",
        "uploads": [],
        "volumes": [],
        "download_time_limit": 1,
        "execution_time_limit": 1,
        "upload_time_limit": 1,
        "streaming_start_time_limit": 1,
    }
    facilitator_signed_fields = SignedFields.from_facilitator_sdk_json(facilitator_request_json)

    v2_job_request = V2JobRequest(
        uuid="uuid",
        executor_class=ExecutorClass.always_on__llm__a6000,
        docker_image="",
        args=[],
        env={},
        use_gpu=False,
        volume=None,
        output_upload=None,
        download_time_limit=1,
        execution_time_limit=1,
        upload_time_limit=1,
        streaming_start_time_limit=1,
    )
    assert v2_job_request.get_signed_fields() == facilitator_signed_fields


def test_signed_fields__volumes_uploads():
    volumes = [
        SingleFileVolume(
            url="smth.amazon.com",
            relative_path="np_data.npy",
        ),
        HuggingfaceVolume(
            repo_id="hug",
            revision="333",
            relative_path="./models/here",
        ),
    ]

    uploads = [
        SingleFilePutUpload(
            url="smth.amazon.com",
            relative_path="output.json",
        )
    ]

    facilitator_request_json = {
        "validator_hotkey": "5HBVrXXYYZZ",
        "random_field": "to_ignore",
        "executor_class": str(ExecutorClass.always_on__llm__a6000),
        "docker_image": "backenddevelopersltd/latest",
        "args": ["--device", "cuda", "--batch_size", "1", "--model_ids", "Deeptensorlab"],
        "env": {"f": "test"},
        "use_gpu": True,
        "input_url": "",
        "uploads": to_json_array(uploads),
        "volumes": to_json_array(volumes),
        "download_time_limit": 1,
        "execution_time_limit": 1,
        "upload_time_limit": 1,
        "streaming_start_time_limit": 1,
    }
    facilitator_signed_fields = SignedFields.from_facilitator_sdk_json(facilitator_request_json)

    v2_job_request = V2JobRequest(
        uuid="uuid",
        executor_class=ExecutorClass.always_on__llm__a6000,
        docker_image="backenddevelopersltd/latest",
        args=["--device", "cuda", "--batch_size", "1", "--model_ids", "Deeptensorlab"],
        env={"f": "test"},
        use_gpu=True,
        volume=MultiVolume(
            volumes=volumes,
        ),
        output_upload=uploads[0],
        download_time_limit=1,
        execution_time_limit=1,
        upload_time_limit=1,
        streaming_start_time_limit=1,
    )
    assert v2_job_request.get_signed_fields() == facilitator_signed_fields


def test_authentication_request(keypair):
    authentication_request = V0AuthenticationRequest.from_keypair(keypair)
    assert authentication_request.verify_signature()
    assert authentication_request.ss58_address == keypair.ss58_address



import abc
import asyncio
import logging
import random
from collections.abc import Awaitable, Callable
from typing import Generic, TypeAlias, TypeVar

from pydantic import BaseModel, ValidationError

from compute_horde.protocol_messages import GenericError
from compute_horde.transport import AbstractTransport, TransportConnectionError

logger = logging.getLogger(__name__)
ErrorCallback: TypeAlias = Callable[[str], Awaitable[None]]

FromMinerType = TypeVar("FromMinerType", bound=BaseModel)
ToMinerType = TypeVar("ToMinerType", bound=BaseModel)


class AbstractMinerClient(Generic[FromMinerType, ToMinerType], metaclass=abc.ABCMeta):
    def __init__(self, miner_name: str, transport: AbstractTransport):
        self.miner_name = miner_name
        self.read_messages_task: asyncio.Task[None] | None = None
        self.deferred_send_tasks: list[asyncio.Task[None]] = []
        self.transport = transport

    @abc.abstractmethod
    def miner_url(self) -> str: ...

    @abc.abstractmethod
    def parse_message(self, raw_msg: str | bytes) -> FromMinerType:
        """Parse raw message into a pydantic model"""

    @abc.abstractmethod
    async def handle_message(self, msg: FromMinerType) -> None:
        """
        Handle the message based on its type or raise UnsupportedMessageReceived
        """
        ...

    async def connect(self) -> None:
        logger.debug(f"Connecting to miner {self.miner_name} at {self.miner_url()}")
        await self.transport.start()
        if self.read_messages_task is None or self.read_messages_task.done():
            self.read_messages_task = asyncio.create_task(self.read_messages())
        logger.debug(f"Connected to miner {self.miner_name} at {self.miner_url()}")

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self) -> None:
        logger.debug(f"Closing connection to miner {self.miner_name} at {self.miner_url()}")
        for deferred_send_task in self.deferred_send_tasks:
            if not deferred_send_task.done():
                deferred_send_task.cancel()
                try:
                    await deferred_send_task
                except asyncio.CancelledError:
                    pass
                except Exception as ex:
                    logger.debug("Exception raised on task cancel: %r", ex)

        if self.read_messages_task is not None and not self.read_messages_task.done():
            self.read_messages_task.cancel()
            try:
                await self.read_messages_task
            except asyncio.CancelledError:
                pass
            except Exception as ex:
                logger.debug("Exception raised on task cancel: %r", ex)

        await self.transport.stop()
        logger.debug(f"Closed connection to miner {self.miner_name} at {self.miner_url()}")

    async def send_model(
        self, model: ToMinerType, error_event_callback: ErrorCallback | None = None
    ) -> None:
        logger.debug(f"Sending model {model} to miner {self.miner_name}")
        await self.send(model.model_dump_json(), error_event_callback)

    async def send(
        self, data: str | bytes, error_event_callback: ErrorCallback | None = None
    ) -> None:
        while True:
            try:
                await self.transport.send(data)
            except TransportConnectionError as ex:
                msg = f"Could not send to miner {self.miner_name}: {str(ex)}"
                logger.warning(msg)
                if error_event_callback:
                    try:
                        await error_event_callback(msg)
                    except Exception as callback_ex:
                        logger.error("Could not execute error event callback: %s", str(callback_ex))
                await asyncio.sleep(1 + random.random())
                continue
            return

    def deferred_send_model(self, model: ToMinerType) -> None:
        task = asyncio.create_task(self.send_model(model))
        self.deferred_send_tasks.append(task)

    async def read_messages(self) -> None:
        async for raw_msg in self.transport:
            try:
                msg = self.parse_message(raw_msg)
            except ValidationError as ex:
                if raw_msg == "PING":
                    continue
                error_msg = f"Malformed message {raw_msg} from miner {self.miner_name}: {ex.json()}"
                logger.info(error_msg)
                self.deferred_send_model(GenericError(details=error_msg))  # type: ignore[arg-type]
                continue

            try:
                await self.handle_message(msg)
            except UnsupportedMessageReceived:
                error_msg = f"Unsupported message from miner {self.miner_name}: {type(msg)}"
                logger.error(error_msg)
                self.deferred_send_model(GenericError(details=error_msg))  # type: ignore[arg-type]


class UnsupportedMessageReceived(Exception):
    def __init__(self, msg: BaseModel):
        self.msg = msg

    def __str__(self):
        return f"{type(self).__name__}: {self.msg.model_dump_json()}"

    __repr__ = __str__



# Generated by Django 5.1.1 on 2024-10-28 12:53

import datetime

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("receipts", "0002_jobstartedreceipt_is_organic"),
    ]

    operations = [
        migrations.CreateModel(
            name="JobAcceptedReceipt",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("job_uuid", models.UUIDField()),
                ("validator_hotkey", models.CharField(max_length=256)),
                ("miner_hotkey", models.CharField(max_length=256)),
                ("validator_signature", models.CharField(max_length=256)),
                ("miner_signature", models.CharField(blank=True, max_length=256, null=True)),
                ("timestamp", models.DateTimeField()),
                ("time_accepted", models.DateTimeField()),
                ("ttl", models.IntegerField()),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.RemoveField(
            model_name="jobstartedreceipt",
            name="time_accepted",
        ),
        migrations.AddField(
            model_name="jobfinishedreceipt",
            name="timestamp",
            field=models.DateTimeField(
                default=datetime.datetime(2020, 1, 1, 0, 0, tzinfo=datetime.UTC)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="jobstartedreceipt",
            name="timestamp",
            field=models.DateTimeField(
                default=datetime.datetime(2020, 1, 1, 0, 0, tzinfo=datetime.UTC)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="jobstartedreceipt",
            name="ttl",
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddIndex(
            model_name="jobfinishedreceipt",
            index=models.Index(fields=["timestamp"], name="jobfinishedreceipt_ts_idx"),
        ),
        migrations.AddIndex(
            model_name="jobstartedreceipt",
            index=models.Index(fields=["timestamp"], name="jobstartedreceipt_ts_idx"),
        ),
        migrations.AddIndex(
            model_name="jobacceptedreceipt",
            index=models.Index(fields=["timestamp"], name="jobacceptedreceipt_ts_idx"),
        ),
        migrations.AddConstraint(
            model_name="jobacceptedreceipt",
            constraint=models.UniqueConstraint(
                fields=("job_uuid",), name="receipts_unique_jobacceptedreceipt_job_uuid"
            ),
        ),
    ]



from compute_horde.receipts.schemas import Receipt, ReceiptType

default_app_config = "compute_horde.receipts.apps.ComputeHordeReceiptsConfig"

# Reexported for compatibility. These were moved to submodules.
__all__ = [
    "Receipt",
    "ReceiptType",
]



import enum
from typing import Annotated, Literal

from compute_horde_core.executor_class import ExecutorClass
from compute_horde_core.output_upload import OutputUpload
from compute_horde_core.streaming import StreamingDetails
from compute_horde_core.volume import Volume
from pydantic import BaseModel, Field

from compute_horde.base.docker import DockerRunOptionsPreset
from compute_horde.receipts.schemas import (
    JobAcceptedReceiptPayload,
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
    Receipt,
)
from compute_horde.utils import MachineSpecs

# NOTE:
# miner.ec = Miner's executor consumer
# miner.vc = Miner's validator consumer


# executor <-> miner.ec <-> miner.vc <-> validator
class GenericError(BaseModel):
    message_type: Literal["GenericError"] = "GenericError"
    details: str | None = None


# validator -> miner.vc
class ValidatorAuthForMiner(BaseModel):
    message_type: Literal["ValidatorAuthForMiner"] = "ValidatorAuthForMiner"
    validator_hotkey: str
    miner_hotkey: str
    timestamp: int
    signature: str

    def blob_for_signing(self) -> str:
        return f"{self.validator_hotkey}:{self.miner_hotkey}:{self.timestamp}"


# miner.vc -> validator
class UnauthorizedError(BaseModel):
    class Code(enum.Enum):
        TOKEN_TOO_OLD = "TOKEN_TOO_OLD"
        UNKNOWN_VALIDATOR = "UNKNOWN_VALIDATOR"
        VALIDATOR_INACTIVE = "VALIDATOR_INACTIVE"

    message_type: Literal["UnauthorizedError"] = "UnauthorizedError"
    code: Code
    details: str | None = None


# miner.vc -> validator
class V0ExecutorManifestRequest(BaseModel):
    message_type: Literal["V0ExecutorManifestRequest"] = "V0ExecutorManifestRequest"
    manifest: dict[ExecutorClass, int]

    @property
    def total_count(self) -> int:
        return sum(self.manifest.values())


# validator -> miner.vc -> miner.ec -> executor
class V0InitialJobRequest(BaseModel):
    class ExecutorTimingDetails(BaseModel):
        allowed_leeway: int
        download_time_limit: int
        execution_time_limit: int
        upload_time_limit: int
        streaming_start_time_limit: int

    message_type: Literal["V0InitialJobRequest"] = "V0InitialJobRequest"
    job_uuid: str
    executor_class: ExecutorClass
    docker_image: str | None = None
    timeout_seconds: int | None = None  # Deprecated - use executor_timing instead
    volume: Volume | None = None
    job_started_receipt_payload: JobStartedReceiptPayload
    job_started_receipt_signature: str

    # this field should be set if the job is a streaming job
    streaming_details: StreamingDetails | None = None

    # This field should be set if the job should use fine-grained timing.
    # Otherwise, the executor will use `timeout_seconds` as the total time limit.
    executor_timing: ExecutorTimingDetails | None = None


# miner.vc -> validator
class V0DeclineJobRequest(BaseModel):
    class Reason(enum.Enum):
        NOT_SPECIFIED = "not_specified"
        BUSY = "busy"
        EXECUTOR_FAILURE = "executor_failure"
        VALIDATOR_BLACKLISTED = "validator_blacklisted"

    message_type: Literal["V0DeclineJobRequest"] = "V0DeclineJobRequest"
    job_uuid: str
    reason: Reason = Reason.NOT_SPECIFIED
    receipts: list[Receipt] = Field(default_factory=list)


# miner.vc -> validator
class V0AcceptJobRequest(BaseModel):
    message_type: Literal["V0AcceptJobRequest"] = "V0AcceptJobRequest"
    job_uuid: str


# executor -> miner.ec -> miner.vc -> validator
class V0ExecutorFailedRequest(BaseModel):
    message_type: Literal["V0ExecutorFailedRequest"] = "V0ExecutorFailedRequest"
    job_uuid: str
    executor_token: str | None = None  # SET ONLY on miner.ec -> miner.vc


# executor -> miner.ec -> miner.vc -> validator
class V0StreamingJobNotReadyRequest(BaseModel):
    message_type: Literal["V0StreamingJobNotReadyRequest"] = "V0StreamingJobNotReadyRequest"
    job_uuid: str
    executor_token: str | None = None  # SET ONLY on miner.ec -> miner.vc


# executor -> miner.ec -> miner.vc -> validator
class V0ExecutorReadyRequest(BaseModel):
    message_type: Literal["V0ExecutorReadyRequest"] = "V0ExecutorReadyRequest"
    job_uuid: str
    executor_token: str | None = None  # SET ONLY on miner.ec -> miner.vc


# signaling the streaming job is ready to accept connections from user
# executor -> miner.ec -> miner.vc -> validator
class V0StreamingJobReadyRequest(BaseModel):
    message_type: Literal["V0StreamingJobReadyRequest"] = "V0StreamingJobReadyRequest"
    job_uuid: str
    executor_token: str | None = None  # SET ONLY on miner.ec -> miner.vc
    public_key: str
    ip: str | None = (
        None  # set by miner after it receives streaming job ready message from executor
    )
    port: int
    miner_signature: str | None = None

    def blob_for_signing(self) -> str:
        return f"{self.job_uuid}:{self.ip}:{self.port}:{self.public_key}"


class V0VolumesReadyRequest(BaseModel):
    message_type: Literal["V0VolumesReadyRequest"] = "V0VolumesReadyRequest"
    job_uuid: str


class V0ExecutionDoneRequest(BaseModel):
    message_type: Literal["V0ExecutionDoneRequest"] = "V0ExecutionDoneRequest"
    job_uuid: str


# validator -> miner.vc -> miner.ec -> executor
class V0JobRequest(BaseModel):
    message_type: Literal["V0JobRequest"] = "V0JobRequest"
    job_uuid: str
    executor_class: ExecutorClass
    docker_image: str
    raw_script: str | None = None
    docker_run_options_preset: DockerRunOptionsPreset
    docker_run_cmd: list[str]
    volume: Volume | None = None
    output_upload: OutputUpload | None = None
    artifacts_dir: str | None = None


# executor -> miner.ec -> miner.vc -> validator
class V0JobFailedRequest(BaseModel):
    class ErrorType(enum.StrEnum):
        TIMEOUT = "TIMEOUT"
        SECURITY_CHECK = "SECURITY_CHECK"
        HUGGINGFACE_DOWNLOAD = "HUGGINGFACE_DOWNLOAD"
        NONZERO_EXIT_CODE = "NONZERO_EXIT_STATUS"
        EXECUTOR_DISCONNECTED = "EXECUTOR_DISCONNECTED"

    message_type: Literal["V0JobFailedRequest"] = "V0JobFailedRequest"
    job_uuid: str
    docker_process_exit_status: int | None = None
    docker_process_stdout: str
    docker_process_stderr: str
    error_type: ErrorType | None = None
    error_detail: str | None = None
    timeout: bool = False


# executor -> miner.ec -> miner.vc -> validator
class V0JobFinishedRequest(BaseModel):
    message_type: Literal["V0JobFinishedRequest"] = "V0JobFinishedRequest"
    job_uuid: str
    docker_process_stdout: str
    docker_process_stderr: str
    artifacts: dict[str, str] | None = None
    upload_results: dict[str, str] | None = (
        None  # Contains serialized HTTP upload results (if available)
    )


# validator -> miner.vc
class V0JobAcceptedReceiptRequest(BaseModel):
    message_type: Literal["V0JobAcceptedReceiptRequest"] = "V0JobAcceptedReceiptRequest"
    payload: JobAcceptedReceiptPayload
    signature: str

    def blob_for_signing(self):
        return self.payload.blob_for_signing()


# validator -> miner.vc
class V0JobFinishedReceiptRequest(BaseModel):
    message_type: Literal["V0JobFinishedReceiptRequest"] = "V0JobFinishedReceiptRequest"
    payload: JobFinishedReceiptPayload
    signature: str

    def blob_for_signing(self):
        return self.payload.blob_for_signing()


# executor -> miner.ec -> miner.vc -> validator
class V0MachineSpecsRequest(BaseModel):
    message_type: Literal["V0MachineSpecsRequest"] = "V0MachineSpecsRequest"
    job_uuid: str
    specs: MachineSpecs


ValidatorToMinerMessage = Annotated[
    GenericError
    | ValidatorAuthForMiner
    | V0InitialJobRequest
    | V0JobRequest
    | V0JobAcceptedReceiptRequest
    | V0JobFinishedReceiptRequest,
    Field(discriminator="message_type"),
]

MinerToExecutorMessage = Annotated[
    GenericError | V0InitialJobRequest | V0JobRequest,
    Field(discriminator="message_type"),
]

ExecutorToMinerMessage = Annotated[
    GenericError
    | V0ExecutorFailedRequest
    | V0StreamingJobNotReadyRequest
    | V0ExecutorReadyRequest
    | V0StreamingJobReadyRequest
    | V0VolumesReadyRequest
    | V0ExecutionDoneRequest
    | V0JobFailedRequest
    | V0JobFinishedRequest
    | V0MachineSpecsRequest,
    Field(discriminator="message_type"),
]

MinerToValidatorMessage = Annotated[
    GenericError
    | UnauthorizedError
    | V0ExecutorManifestRequest
    | V0DeclineJobRequest
    | V0AcceptJobRequest
    | V0ExecutorFailedRequest
    | V0StreamingJobNotReadyRequest
    | V0ExecutorReadyRequest
    | V0StreamingJobReadyRequest
    | V0JobFailedRequest
    | V0VolumesReadyRequest
    | V0ExecutionDoneRequest
    | V0JobFinishedRequest
    | V0MachineSpecsRequest,
    Field(discriminator="message_type"),
]



import abc
import datetime
from collections.abc import Sequence

from compute_horde.receipts.schemas import Receipt


class BaseReceiptStore(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def store(self, receipts: Sequence[Receipt]) -> None:
        """
        Append receipts to the store.
        """
        ...

    @abc.abstractmethod
    def evict(self, cutoff: datetime.datetime) -> None:
        """
        Remove receipts (roughly) older than the cutoff
        """
        ...



from compute_horde.settings import *  # noqa

USE_TZ = True

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

INSTALLED_APPS += [
    "tests.test_app",
]

CELERY_TASK_ALWAYS_EAGER = True

COMPUTE_HORDE_BLOCK_CACHE_KEY = "test-block-cache-key"

COMPUTE_HORDE_VALIDATOR_MODEL = "test_app.Validator"
COMPUTE_HORDE_VALIDATOR_KEY_FIELD = "public_key"
COMPUTE_HORDE_VALIDATOR_ACTIVE_FIELD = "active"
COMPUTE_HORDE_VALIDATOR_DEBUG_FIELD = "debug"

BITTENSOR_NETUID = 49
BITTENSOR_NETWORK = "local"



import asyncio
import functools

import bittensor
from asgiref.sync import async_to_sync
from django.conf import settings
from django.core.cache import cache

_BLOCK_CACHE_KEY = getattr(
    settings, "COMPUTE_HORDE_BLOCK_CACHE_KEY", "compute_horde.blockchain.block_cache.current_block"
)
_BLOCK_CACHE_TIMEOUT = getattr(settings, "COMPUTE_HORDE_BLOCK_CACHE_TIMEOUT", 2)


class BlockNotInCacheError(KeyError):
    pass


@functools.cache
def _get_subtensor(network):
    return bittensor.subtensor(network)


async def aget_current_block(timeout: float = 1.0) -> int:
    """
    Gets the current block number from cache. Waits for ``timeout`` seconds if it's not there.

    :param timeout: Number of seconds to wait for the block. Pass 0 to return or raise an exception immediately instead.
    :raise BlockNotInCacheError: If block was not found in cache.
    """
    # Note: cache.get here is not async.
    current_block: int = cache.get(_BLOCK_CACHE_KEY)
    if current_block is not None:
        return current_block

    if timeout <= 0:
        raise BlockNotInCacheError(_BLOCK_CACHE_KEY)

    await asyncio.sleep(timeout)

    return await aget_current_block(timeout=0)


get_current_block = async_to_sync(aget_current_block)


def cache_current_block() -> None:
    subtensor = _get_subtensor(network=settings.BITTENSOR_NETWORK)
    current_block = subtensor.get_current_block()

    cache.set(_BLOCK_CACHE_KEY, current_block, _BLOCK_CACHE_TIMEOUT)



from django.apps import AppConfig


class ComputeHordeBlockchainConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "compute_horde.blockchain"
    verbose_name = "Blockchain"

    def ready(self):
        from django.conf import settings

        # Check if a required setting is defined
        if not hasattr(settings, "BITTENSOR_NETWORK"):
            raise ValueError(
                "The setting 'BITTENSOR_NETWORK' is required in your settings.py "
                "for 'compute_horde.blockchain' to function properly. Please define it."
            )



import datetime
import enum
import json
from typing import Annotated, Literal

import bittensor
from compute_horde_core.executor_class import ExecutorClass
from pydantic import BaseModel, Field


class ReceiptType(enum.StrEnum):
    JobStartedReceipt = "JobStartedReceipt"
    JobAcceptedReceipt = "JobAcceptedReceipt"
    JobFinishedReceipt = "JobFinishedReceipt"


class BaseReceiptPayload(BaseModel):
    job_uuid: str
    miner_hotkey: str
    validator_hotkey: str
    timestamp: datetime.datetime  # when the receipt was generated

    def blob_for_signing(self):
        # pydantic v2 does not support sort_keys anymore.
        return json.dumps(self.model_dump(mode="json"), sort_keys=True)


class JobStartedReceiptPayload(BaseReceiptPayload):
    receipt_type: Literal[ReceiptType.JobStartedReceipt] = ReceiptType.JobStartedReceipt
    executor_class: ExecutorClass
    is_organic: bool
    ttl: int


class JobAcceptedReceiptPayload(BaseReceiptPayload):
    receipt_type: Literal[ReceiptType.JobAcceptedReceipt] = ReceiptType.JobAcceptedReceipt
    time_accepted: datetime.datetime
    ttl: int  # unused


class JobFinishedReceiptPayload(BaseReceiptPayload):
    receipt_type: Literal[ReceiptType.JobFinishedReceipt] = ReceiptType.JobFinishedReceipt
    time_started: datetime.datetime
    time_took_us: int  # micro-seconds
    score_str: str

    @property
    def time_took(self):
        return datetime.timedelta(microseconds=self.time_took_us)

    @property
    def score(self):
        return float(self.score_str)


ReceiptPayload = Annotated[
    JobStartedReceiptPayload | JobAcceptedReceiptPayload | JobFinishedReceiptPayload,
    Field(discriminator="receipt_type"),
]


class BadReceiptSignature(Exception):
    def __init__(self, receipt: "Receipt"):
        self.receipt = receipt


class BadMinerReceiptSignature(BadReceiptSignature):
    pass


class BadValidatorReceiptSignature(BadReceiptSignature):
    pass


class Receipt(BaseModel):
    payload: ReceiptPayload
    validator_signature: str
    miner_signature: str

    def verify_miner_signature(self, throw=False):
        miner_keypair = bittensor.Keypair(ss58_address=self.payload.miner_hotkey)
        is_valid = miner_keypair.verify(self.payload.blob_for_signing(), self.miner_signature)
        if throw and not is_valid:
            raise BadMinerReceiptSignature(self)
        return is_valid

    def verify_validator_signature(self, throw=False):
        validator_keypair = bittensor.Keypair(ss58_address=self.payload.validator_hotkey)
        blob = self.payload.blob_for_signing()
        try:
            is_valid = validator_keypair.verify(blob, self.validator_signature)
        except Exception:
            is_valid = False
        if throw and not is_valid:
            raise BadValidatorReceiptSignature(self)
        return is_valid



# Generated by Django 4.2.16 on 2024-10-15 13:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("receipts", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="jobstartedreceipt",
            name="is_organic",
            field=models.BooleanField(default=False),
            preserve_default=False,
        ),
    ]



from collections.abc import Iterable
from unittest.mock import MagicMock, patch

import pytest
from faker import Faker

from compute_horde.validators.tasks import fetch_validators
from tests.test_app.models import Validator, Validator2


@pytest.fixture(autouse=True)
def mock_get_validators():
    with patch("compute_horde.validators.tasks.get_validators") as p:
        yield p


@pytest.fixture
def active_validators_keys(faker: Faker):
    return [faker.pystr() for _ in range(5)]


@pytest.fixture
def inactive_validators_keys(faker: Faker):
    return [faker.pystr() for _ in range(4)]


def _create_validators(keys: Iterable[str]):
    validator_models = [
        (Validator, "public_key", "active"),
        (Validator2, "ss58_address", "is_active"),
    ]
    for model, key_field, active_field in validator_models:
        model.objects.bulk_create([model(**{key_field: key, active_field: True}) for key in keys])


@pytest.fixture(autouse=True)
def active_validators(active_validators_keys: list[str]):
    _create_validators(active_validators_keys)


@pytest.fixture(autouse=True)
def inactive_validators(inactive_validators_keys: list[str]):
    _create_validators(inactive_validators_keys)


def _generate_validator_mocks(hotkeys: Iterable[str]):
    return [MagicMock(hotkey=key) for key in hotkeys]


@pytest.mark.django_db(transaction=True)
def test_fetch_validators_creates_new_active_validators(
    faker: Faker, mock_get_validators: MagicMock
):
    new_keys = [faker.pystr() for _ in range(3)]
    mock_get_validators.return_value = _generate_validator_mocks(new_keys)

    fetch_validators()

    for key in new_keys:
        assert Validator.objects.filter(public_key=key, active=True).exists()


@pytest.mark.django_db(transaction=True)
def test_fetch_validators_creates_new_active_validators__different_fields(
    faker: Faker, mock_get_validators: MagicMock, settings
):
    settings.COMPUTE_HORDE_VALIDATOR_MODEL = "test_app.Validator2"
    settings.COMPUTE_HORDE_VALIDATOR_KEY_FIELD = "ss58_address"
    settings.COMPUTE_HORDE_VALIDATOR_ACTIVE_FIELD = "is_active"
    settings.COMPUTE_HORDE_VALIDATOR_DEBUG_FIELD = None

    new_keys = [faker.pystr() for _ in range(3)]
    mock_get_validators.return_value = _generate_validator_mocks(new_keys)

    fetch_validators()

    for key in new_keys:
        assert Validator2.objects.filter(ss58_address=key, is_active=True).exists()

        # they should not end up in the wrong model
        assert not Validator.objects.filter(public_key=key, active=True).exists()


@pytest.mark.django_db(transaction=True)
def test_fetch_validators_updates_existing_validators(
    active_validators_keys: list[str],
    inactive_validators_keys: list[str],
    mock_get_validators: MagicMock,
):
    mock_get_validators.return_value = _generate_validator_mocks(inactive_validators_keys)

    fetch_validators()

    assert Validator.objects.count() == len(active_validators_keys) + len(inactive_validators_keys)

    assert Validator.objects.filter(active=True).count() == len(inactive_validators_keys)

    for key in inactive_validators_keys:
        assert Validator.objects.filter(public_key=key, active=True).exists()


@pytest.mark.django_db(transaction=True)
def test_fetch_validators_debug(
    active_validators_keys: list[str],
    inactive_validators_keys: list[str],
    mock_get_validators: MagicMock,
    faker: Faker,
):
    mock_get_validators.return_value = _generate_validator_mocks(active_validators_keys)

    debug_validator_key = faker.pystr()
    Validator.objects.create(public_key=debug_validator_key, active=True, debug=True)

    fetch_validators()

    assert (
        Validator.objects.count() == len(active_validators_keys) + len(inactive_validators_keys) + 1
    )
    assert set(Validator.objects.filter(active=True).values_list("public_key", flat=True)) == set(
        active_validators_keys
    ) | {debug_validator_key}


@pytest.mark.django_db(transaction=True)
def test_fetch_validators_debug_inactive(
    active_validators_keys: list[str],
    mock_get_validators: MagicMock,
    faker: Faker,
):
    mock_get_validators.return_value = _generate_validator_mocks(active_validators_keys)

    debug_validator_key = faker.pystr()
    debug_validator = Validator.objects.create(
        public_key=debug_validator_key, active=False, debug=True
    )

    fetch_validators()

    debug_validator.refresh_from_db()
    assert not debug_validator.active



import asyncio
import logging
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from io import BytesIO
from typing import TypeAlias

import aiohttp
import pydantic
from aiohttp import ClientTimeout
from asgiref.sync import sync_to_async
from django.core.exceptions import ImproperlyConfigured

from compute_horde.receipts.models import ReceiptModel, receipt_to_django_model
from compute_horde.receipts.schemas import (
    BadMinerReceiptSignature,
    BadValidatorReceiptSignature,
    Receipt,
)
from compute_horde.receipts.transfer_checkpoints import checkpoint_backend

logger = logging.getLogger(__name__)

MinerInfo: TypeAlias = tuple[str, str, int]


class TransferException(Exception):
    """
    Transfer exception means the request for receipts completely failed - timed out, got an error response etc.
    """

    def __init__(self, miner: MinerInfo, page: int):
        super().__init__(f"Transfer failed: {miner=} {page=}")
        self.miner: MinerInfo = miner
        self.page: int = page


class LineException(Exception):
    """
    A line exception means that, while processing a correct response, one of the receipt lines could not be processed.
    """

    def __init__(
        self,
        cause: pydantic.ValidationError | BadMinerReceiptSignature | BadValidatorReceiptSignature,
    ):
        super().__init__(f"Transferred line skipped: {type(cause).__name__}: {cause}")
        self.__cause__ = self.cause = cause


@dataclass
class TransferResult:
    n_receipts: int
    n_successful_transfers: int
    transfer_errors: list[TransferException]
    line_errors: list[LineException]


class ReceiptsTransfer:
    """
    HTTP client for fetching receipts from a target based on an HTTP URL.
    Uses HTTP range requests to only retrieve receipts added since last run.
    """

    @classmethod
    async def transfer(
        cls,
        miners: Sequence[MinerInfo],
        pages: Sequence[int],
        session: aiohttp.ClientSession,
        semaphore: asyncio.Semaphore,
        request_timeout: float,
    ) -> TransferResult:
        """
        Efficiently transfer receipts from multiple miners at the same time, storing them in local database.
        Still, the number of pages transferred at the same time should be limited to 1-2.
        """

        async def rate_limited_transfer(
            transfer: ReceiptsTransfer, page: int, session: aiohttp.ClientSession
        ) -> tuple[list[Receipt], list[LineException]]:
            try:
                async with semaphore:
                    # This both fetches and verifies the receipts.
                    return await transfer.get_new_receipts_on_page(
                        page=page,
                        session=session,
                        timeout=request_timeout,
                    )
            except (RuntimeError, ImproperlyConfigured):
                # Don't catch this as it may block the script from exiting.
                raise
            except (TimeoutError, Exception) as e:
                # Wrap other exceptions as transfer exceptions
                raise TransferException(miner, page) from e

        # Create transfer tasks, one transfer task = one page from one miner
        transfer_tasks = []
        for page in pages:
            for miner in miners:
                hotkey, ip, port = miner
                transfer = cls(f"http://{ip}:{port}/receipts")
                transfer_tasks.append(
                    asyncio.create_task(rate_limited_transfer(transfer, page, session))
                )

        # Place received receipts into buckets based on receipt model type
        receipts_by_type: defaultdict[type[ReceiptModel], list[ReceiptModel]] = defaultdict(list)
        total_receipts = 0
        successful_transfers = 0
        transfer_errors: list[TransferException] = []
        line_errors: list[LineException] = []

        # Wait for transfer tasks in parallel, handle a batch of receipts as soon as any is available
        for transfer_task in asyncio.as_completed(transfer_tasks):
            try:
                transferred_batch, batch_line_errors = await transfer_task
                line_errors.extend(batch_line_errors)
                successful_transfers += 1
                for receipt in transferred_batch:
                    model = receipt_to_django_model(receipt)
                    model_type = type(model)
                    bucket = receipts_by_type[model_type]
                    bucket.append(model)
                    if len(bucket) >= 1000:
                        await model_type.objects.abulk_create(bucket, ignore_conflicts=True)  # type: ignore
                        total_receipts += len(bucket)
                        logger.info(f"Stored {len(bucket)} {model_type.__name__} receipts")
                        bucket.clear()
            except TransferException as e:
                transfer_errors.append(e)
                continue

        # Insert the remainder of the receipts
        for model_type, bucket in receipts_by_type.items():
            await model_type.objects.abulk_create(bucket, ignore_conflicts=True)  # type: ignore
            total_receipts += len(bucket)
            logger.info(f"Stored {len(bucket)} {model_type.__name__} receipts")

        return TransferResult(total_receipts, successful_transfers, transfer_errors, line_errors)

    def __init__(self, server_url: str):
        self._receipts_url = server_url.rstrip("/")

    async def get_new_receipts_on_page(
        self, page: int, session: aiohttp.ClientSession, timeout: float
    ) -> tuple[list[Receipt], list[LineException]]:
        """
        Fetch a batch of receipts from remote server.
        Will start from last checkpoint for this page, if available.
        """
        page_url = self.page_url(page)
        checkpoints = checkpoint_backend()

        checkpoint_key = page_url
        checkpoint = await checkpoints.get(checkpoint_key)
        use_range_request = checkpoint > 0

        if use_range_request:
            # We're re-requesting the page starting from a known offset.
            # Ask for file content that was added after the page was last checked.
            # Also, range request and gzip won't work together.
            # (the range relates to compressed bytes then, which are meaningless here.)
            headers = {
                "Accept-Encoding": "",
                "Range": f"bytes={checkpoint}-",
            }
        else:
            # This is the first time we're requesting this page.
            # This will only receive a gzipped page if it's available.
            # If it's not, this will still work but the raw page file will be sent.
            # aiohttp inflates the file automatically.
            headers = {
                "Accept-Encoding": "gzip",
            }

        # As the request should be as fast as possible, don't allow redirecting - clients must respond immediately.
        response = await session.get(
            page_url, headers=headers, allow_redirects=False, timeout=ClientTimeout(total=timeout)
        )
        if response.status in {404, 416}:
            # 404 - miner doesn't have the page (yet / anymore / at all)
            # 416 - no new receipts on page
            logger.debug(f"Nothing to fetch from {page_url}: {response.status}")
            return [], []
        if response.status not in {200, 206}:
            raise Exception(f"Request failed for {page_url}: {response.status}")
        jsonl_content = await response.read()

        # Put this on a worker thread, otherwise other HTTP requests are more likely to time out waiting for this.
        receipts, line_errors = await sync_to_async(
            self._to_valid_receipts, thread_sensitive=False
        )(jsonl_content)

        # Save the checkpoint
        if use_range_request:
            # Range requests return a "content-range" header that contains full size of the file.
            # We ask the server not to gzip range requests - so this is the total bytes we got so far.
            range_header = response.headers["content-range"]
            assert range_header.lower().startswith("bytes ")
            _, total_str = range_header.split("/")
            await checkpoints.set(checkpoint_key, int(total_str))
        else:
            # We got the complete page file, use its size for next checkpoint.
            await checkpoints.set(checkpoint_key, len(jsonl_content))

        return receipts, line_errors

    def page_url(self, page: int) -> str:
        return f"{self._receipts_url}/{page}.jsonl"

    def _to_valid_receipts(
        self, received_bytes: bytes
    ) -> tuple[list[Receipt], list[LineException]]:
        """
        Converts a received JSONL chunk into valid, signed receipts - one line at a time.
        Skips anything else.
        """
        receipts: list[Receipt] = []
        line_errors: list[LineException] = []
        for line in BytesIO(received_bytes):
            try:
                receipt = Receipt.model_validate_json(line)
                receipt.verify_miner_signature(throw=True)
                receipt.verify_validator_signature(throw=True)
                receipts.append(receipt)
            except pydantic.ValidationError as e:
                # This is potentially a serious issue.
                # A receipt line that doesn't validate could be a schema mismatch or an incomplete receipt line.
                # Schema incompatibility is a showstopper for syncing a miner.
                # Incomplete receipts will get lost in transfer and the cause of it must be investigated.
                logger.info(
                    "skipping line: failed validation: %s%s",
                    line[:100],
                    " (...)" if len(line) > 100 else "",
                )
                line_errors.append(LineException(e))
                continue
            except BadMinerReceiptSignature as e:
                logger.info(
                    "Skipping %s with bad miner signature: %s",
                    e.receipt.payload.receipt_type,
                    e.receipt.payload.job_uuid,
                )
                line_errors.append(LineException(e))
                continue
            except BadValidatorReceiptSignature as e:
                logger.info(
                    "Skipping %s with bad validator signature: %s",
                    e.receipt.payload.receipt_type,
                    e.receipt.payload.job_uuid,
                )
                line_errors.append(LineException(e))
                continue
        return receipts, line_errors



import asyncio
from unittest.mock import MagicMock, patch

import pytest
from celery import Celery
from django.conf import settings
from django.core.cache import cache

from compute_horde.blockchain.block_cache import (
    BlockNotInCacheError,
    aget_current_block,
    get_current_block,
)
from compute_horde.blockchain.tasks import update_block_cache


@pytest.fixture(autouse=True)
def clear_cache():
    cache.clear()
    yield


@pytest.fixture
def celery():
    app = Celery("test")
    app.config_from_object("django.conf:settings", namespace="CELERY")
    yield app
    app.close()


@pytest.mark.asyncio
async def test_block_cache_aget_current_block_in_cache():
    cache.set(settings.COMPUTE_HORDE_BLOCK_CACHE_KEY, 123)

    current_block = await aget_current_block()

    assert current_block == 123


@pytest.mark.asyncio
async def test_block_cache_aget_current_block_wait():
    async def populate_after_delay():
        await asyncio.sleep(0.1)
        cache.set(settings.COMPUTE_HORDE_BLOCK_CACHE_KEY, 123)

    asyncio.create_task(populate_after_delay())

    current_block = await aget_current_block()

    assert current_block == 123


@pytest.mark.asyncio
async def test_block_cache_aget_current_block_timeout():
    with pytest.raises(BlockNotInCacheError):
        await aget_current_block(timeout=0.01)


def test_block_cache_get_current_block_in_cache():
    cache.set(settings.COMPUTE_HORDE_BLOCK_CACHE_KEY, 123)

    current_block = get_current_block()

    assert current_block == 123


def test_block_cache_get_current_block_timeout():
    with pytest.raises(BlockNotInCacheError):
        get_current_block()


def test_tasks_update_block_cache(celery):
    mock_subtensor = MagicMock()
    with patch("bittensor.subtensor", return_value=mock_subtensor):
        mock_subtensor.get_current_block.return_value = 123

        update_block_cache.delay()

        current_block = get_current_block()

        assert current_block == 123

        mock_subtensor.get_current_block.assert_called_once()



from asyncio import Semaphore
from functools import partial
from unittest.mock import AsyncMock, Mock, patch

import aiohttp
import pytest
from pydantic import ValidationError

from compute_horde.receipts.models import JobStartedReceipt
from compute_horde.receipts.transfer import ReceiptsTransfer
from compute_horde.receipts.transfer_checkpoints import TransferCheckpointBackend
from compute_horde.utils import sign_blob
from tests.utils import random_keypair, random_receipt

do_transfer = partial(
    ReceiptsTransfer.transfer,
    miners=[("miner-hotkey", "ip-address", 12345)],
    pages=[1],
    semaphore=Semaphore(100),
    request_timeout=100,
)


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_rejects_bad_lines(miner_keypair, validator_keypair):
    lines_to_serve = [
        random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
        b"not a receipt",
        b"",
        b"",
        random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
        random_receipt(miner_keypair, validator_keypair).model_dump_json().encode()[50:],
        b"garbage",
        random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
    ]

    session = Mock()
    session.get = AsyncMock()
    session.get.return_value.status = 200
    session.get.return_value.read = AsyncMock(return_value=b"\n".join(lines_to_serve))

    result = await do_transfer(session=session)

    assert result.n_receipts == 3
    assert len(result.transfer_errors) == 0
    assert len(result.line_errors) == 5
    assert await JobStartedReceipt.objects.acount() == 3


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_rejects_bad_signatures(miner_keypair, validator_keypair):
    bad_miner_signature = random_receipt(miner_keypair, validator_keypair)
    bad_miner_signature.miner_signature = sign_blob(
        random_keypair(),
        bad_miner_signature.payload.blob_for_signing(),
    )

    bad_validator_signature = random_receipt(miner_keypair, validator_keypair)
    bad_validator_signature.validator_signature = sign_blob(
        random_keypair(),
        bad_validator_signature.payload.blob_for_signing(),
    )

    both_bad_signatures = random_receipt(miner_keypair, validator_keypair)
    both_bad_signatures.miner_signature = sign_blob(
        random_keypair(),
        both_bad_signatures.payload.blob_for_signing(),
    )
    both_bad_signatures.validator_signature = sign_blob(
        random_keypair(),
        both_bad_signatures.payload.blob_for_signing(),
    )

    lines_to_serve = [
        bad_miner_signature.model_dump_json().encode(),
        bad_validator_signature.model_dump_json().encode(),
        both_bad_signatures.model_dump_json().encode(),
    ]

    session = Mock()
    session.get = AsyncMock()
    session.get.return_value.status = 200
    session.get.return_value.read = AsyncMock(return_value=b"\n".join(lines_to_serve))

    result = await do_transfer(session=session)

    assert result.n_receipts == 0
    assert len(result.transfer_errors) == 0
    assert len(result.line_errors) == 3
    assert await JobStartedReceipt.objects.acount() == 0


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_handles_failures(miner_keypair, mocked_checkpoint_backend):
    session = Mock()

    mock_response_500 = Mock()
    mock_response_500.status = 500
    mock_response_400 = Mock()
    mock_response_400.status = 400

    response_effects = [
        Exception("Doh!"),
        ValueError("Doh!"),
        ValidationError("Doh!", []),
        aiohttp.ClientError("Doh!"),
        TimeoutError("Doh!"),
        mock_response_500,
        mock_response_400,
    ]
    session.get = AsyncMock(side_effect=response_effects)

    result = await do_transfer(session=session, pages=[*range(len(response_effects))])

    assert result.n_receipts == 0
    assert len(result.transfer_errors) == len(response_effects)
    assert len(result.line_errors) == 0
    assert mocked_checkpoint_backend.value == 0
    assert await JobStartedReceipt.objects.acount() == 0
    # implicit assertion - none of the exceptions sneak out


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_uses_checkpoints(miner_keypair, validator_keypair, mocked_checkpoint_backend):
    session = Mock()
    session.get = AsyncMock()

    # First response: whole page
    initial_data = b"\n".join(
        [
            random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
            random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
            random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
        ]
    )
    initial_response = Mock()
    initial_response.read = AsyncMock(return_value=initial_data)
    initial_response.status = 200

    # Subsequent response: changes on top of first page
    # This should be a range request with offset == length of previous response
    subsequent_data = b"\n".join(
        [
            random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
            random_receipt(miner_keypair, validator_keypair).model_dump_json().encode(),
        ]
    )
    subsequent_response = Mock()
    subsequent_response.status = 206
    subsequent_response.read = AsyncMock(return_value=subsequent_data)
    subsequent_response.headers = {"content-range": "bytes doesntmatter/1337"}

    # Final response: nothing more to serve
    # This should also be a range request with offset == 1337 - taken from the content-range header of previous response
    no_more_content_response = Mock()
    no_more_content_response.status = 416
    no_more_content_response.read = AsyncMock(return_value=b"")

    session.get.side_effect = [
        initial_response,
        subsequent_response,
        no_more_content_response,
    ]

    result = await do_transfer(session=session)
    assert result.n_receipts == 3
    assert len(result.transfer_errors) == 0
    assert len(result.line_errors) == 0
    assert mocked_checkpoint_backend.value == len(initial_data)
    assert await JobStartedReceipt.objects.acount() == 3

    result = await do_transfer(session=session)
    _, get_kwargs = session.get.call_args
    assert get_kwargs["headers"]["Range"] == f"bytes={len(initial_data)}-"
    assert result.n_receipts == 2
    assert len(result.transfer_errors) == 0
    assert len(result.line_errors) == 0
    assert mocked_checkpoint_backend.value == 1337
    assert await JobStartedReceipt.objects.acount() == 5

    result = await do_transfer(session=session)
    _, get_kwargs = session.get.call_args
    assert get_kwargs["headers"]["Range"] == f"bytes={1337}-"
    assert result.n_receipts == 0
    assert len(result.transfer_errors) == 0
    assert len(result.line_errors) == 0
    assert mocked_checkpoint_backend.value == 1337
    assert await JobStartedReceipt.objects.acount() == 5


@pytest.fixture(autouse=True)
def mocked_checkpoint_backend():
    """
    Creates an in-memory single-value checkpoint backend for each test.
    """
    with patch("compute_horde.receipts.transfer.checkpoint_backend") as mocked:
        backend = MockTransferCheckpointBackend()
        mocked.return_value = backend
        yield backend


class MockTransferCheckpointBackend(TransferCheckpointBackend):
    def __init__(self):
        self.value = 0

    async def get(self, key: str) -> int:
        return self.value

    async def set(self, key: str, checkpoint: int) -> None:
        self.value = checkpoint



from enum import StrEnum
from typing import Any, Literal, Self

import bittensor
from pydantic import BaseModel


class V0Heartbeat(BaseModel, extra="forbid"):
    """Message sent from validator to facilitator to keep connection alive"""

    message_type: Literal["V0Heartbeat"] = "V0Heartbeat"


class V0AuthenticationRequest(BaseModel, extra="forbid"):
    """Message sent from validator to facilitator to authenticate itself"""

    message_type: Literal["V0AuthenticationRequest"] = "V0AuthenticationRequest"
    public_key: str
    signature: str

    @classmethod
    def from_keypair(cls, keypair: bittensor.Keypair) -> Self:
        return cls(
            public_key=keypair.public_key.hex(),
            signature=f"0x{keypair.sign(keypair.public_key).hex()}",
        )

    def verify_signature(self) -> bool:
        public_key_bytes = bytes.fromhex(self.public_key)
        keypair = bittensor.Keypair(public_key=self.public_key, ss58_format=42)
        # make mypy happy
        valid: bool = keypair.verify(public_key_bytes, self.signature)
        return valid

    @property
    def ss58_address(self) -> str:
        # make mypy happy
        address: str = bittensor.Keypair(public_key=self.public_key, ss58_format=42).ss58_address
        return address


class MinerResponse(BaseModel, extra="allow"):
    job_uuid: str
    message_type: str | None
    docker_process_stderr: str
    docker_process_stdout: str
    artifacts: dict[str, str] | None = None
    upload_results: dict[str, str] | None = None


class StreamingServerDetails(BaseModel, extra="forbid"):
    streaming_server_cert: str | None = None
    streaming_server_address: str | None = None
    streaming_server_port: int | None = None


class JobStatusMetadata(BaseModel, extra="allow"):
    comment: str
    miner_response: MinerResponse | None = None
    streaming_details: StreamingServerDetails | None = None


class JobStatusUpdate(BaseModel, extra="forbid"):
    """
    Message sent from validator to facilitator in response to NewJobRequest.
    """

    class Status(StrEnum):
        RECEIVED = "received"
        ACCEPTED = "accepted"
        EXECUTOR_READY = "executor_ready"
        VOLUMES_READY = "volumes_ready"
        EXECUTION_DONE = "execution_done"
        COMPLETED = "completed"
        REJECTED = "rejected"
        FAILED = "failed"
        STREAMING_READY = "streaming_ready"

    message_type: Literal["V0JobStatusUpdate"] = "V0JobStatusUpdate"
    uuid: str
    status: Status
    metadata: JobStatusMetadata | None = None


class V0MachineSpecsUpdate(BaseModel, extra="forbid"):
    """Message sent from validator to facilitator to update miner specs"""

    message_type: Literal["V0MachineSpecsUpdate"] = "V0MachineSpecsUpdate"
    miner_hotkey: str
    validator_hotkey: str
    specs: dict[str, Any]
    batch_id: str



# Generated by Django 4.2.19 on 2025-04-27 23:22

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("receipts", "0003_jobacceptedreceipt_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="jobstartedreceipt",
            name="max_timeout",
        ),
    ]



import asyncio
import dataclasses
import datetime
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import Any, ParamSpec, TypeVar

import bittensor
import pydantic
from bittensor.core.errors import SubstrateRequestException

try:
    import turbobt
except ImportError:
    turbobt = None

BAC_VALIDATOR_SS58_ADDRESS = "5HBVrFGy6oYhhh71m9fFGYD7zbKyAeHnWN8i8s9fJTBMCtEE"
MIN_VALIDATOR_STAKE = 1000
VALIDATORS_LIMIT = 24


class MachineSpecs(pydantic.BaseModel):
    specs: dict[Any, Any]

    def __str__(self) -> str:
        return str(self.specs)


class ValidatorListError(Exception):
    def __init__(self, reason: Exception):
        self.reason = reason


@dataclasses.dataclass
class ValidatorInfo:
    uid: int
    hotkey: str
    stake: float  # total effective stake denominated in alpha tokens


def get_validators(
    metagraph: bittensor.Metagraph | None = None,
    netuid=12,
    network="finney",
    block: int | None = None,
) -> list[ValidatorInfo]:
    """
    Validators are top 24 neurons in terms of stake, only taking into account those that have at least 1000
    and forcibly including BAC_VALIDATOR_SS58_ADDRESS.
    The result is sorted.
    """
    if metagraph is None:
        try:
            subtensor = bittensor.subtensor(network=network)
        except Exception as ex:
            raise ValidatorListError(ex) from ex

        try:
            metagraph = subtensor.metagraph(netuid, block=block)
        except SubstrateRequestException as ex:
            raise ValidatorListError(ex) from ex

    neurons = [
        n
        for n in metagraph.neurons
        if (
            n.hotkey == BAC_VALIDATOR_SS58_ADDRESS
            or metagraph.total_stake[n.uid] >= MIN_VALIDATOR_STAKE
        )
    ]
    neurons = sorted(
        neurons,
        key=lambda n: (n.hotkey == BAC_VALIDATOR_SS58_ADDRESS, metagraph.total_stake[n.uid]),
        reverse=True,
    )
    return [
        ValidatorInfo(uid=n.uid, hotkey=n.hotkey, stake=float(metagraph.total_stake[n.uid]))
        for n in neurons[:VALIDATORS_LIMIT]
    ]


async def turbobt_get_validators(
    bittensor: "turbobt.Bittensor",
    netuid=12,
    block: int | None = None,
) -> list["turbobt.Neuron"]:
    """
    Validators are top 64 neurons in terms of stake, only taking into account those that have at least 1000
    and forcibly including BAC_VALIDATOR_SS58_ADDRESS.
    The result is sorted.
    """
    if turbobt is None:
        raise ImportError("turbobt")

    subnet = bittensor.subnet(netuid)

    async with bittensor.blocks[block]:
        validators: list[turbobt.Neuron] = await subnet.list_validators()
        validators.sort(
            key=lambda validator: (
                validator.hotkey == BAC_VALIDATOR_SS58_ADDRESS,
                validator.stake,
            ),
            reverse=True,
        )

        if not validators or validators[0].hotkey != BAC_VALIDATOR_SS58_ADDRESS:
            validator = await subnet.get_neuron(BAC_VALIDATOR_SS58_ADDRESS)

            if validator:
                validators.insert(0, validator)

        return validators


def json_dumps_default(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()

    raise TypeError


class Timer:
    def __init__(self, timeout: float | None = None) -> None:
        self.start_time = datetime.datetime.now()
        self.timeout = timeout

    def set_timeout(self, seconds: float) -> None:
        self.start_time = datetime.datetime.now()
        self.timeout = seconds

    def extend_timeout(self, seconds: float) -> None:
        if self.timeout is None:
            raise ValueError("timeout was not specified")
        self.timeout += seconds

    def passed_time(self) -> float:
        return (datetime.datetime.now() - self.start_time).total_seconds()

    def time_left(self) -> float:
        if self.timeout is None:
            raise ValueError("timeout was not specified")
        return self.timeout - self.passed_time()


def sign_blob(kp: bittensor.Keypair, blob: str) -> str:
    """
    Signs a string blob with a bittensor keypair and returns the signature
    """
    return f"0x{kp.sign(blob).hex()}"


P = ParamSpec("P")
R = TypeVar("R")


def async_synchronized(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
    """
    Wraps the function in an async lock.
    """
    lock = asyncio.Lock()

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        async with lock:
            return await func(*args, **kwargs)

    return wrapper



# Generated by Django 4.2.16 on 2025-01-02 08:55

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Validator",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("public_key", models.TextField(unique=True)),
                ("active", models.BooleanField()),
                ("debug", models.BooleanField(default=False)),
            ],
        ),
    ]



import logging

from celery import shared_task

from .block_cache import cache_current_block

logger = logging.getLogger(__name__)


@shared_task()
def update_block_cache():
    logger.info("Updating block cache")
    cache_current_block()



# Generated by Django 4.2.19 on 2025-07-08 19:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Validator2",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("ss58_address", models.CharField(max_length=48, unique=True)),
                ("is_active", models.BooleanField()),
            ],
        ),
    ]



import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class ComputeHordeReceiptsConfig(AppConfig):
    name = "compute_horde.receipts"
    verbose_name = "Receipts"
    default_auto_field = "django.db.models.BigAutoField"



import datetime
from collections.abc import Sequence

from compute_horde.receipts import Receipt
from compute_horde.receipts.store.base import BaseReceiptStore


class NoopReceiptStore(BaseReceiptStore):
    def store(self, receipts: Sequence[Receipt]) -> None:
        pass

    def evict(self, cutoff: datetime.datetime) -> None:
        pass



import asyncio
import contextlib
import datetime
import enum
import logging
import time
from dataclasses import asdict, dataclass, field
from functools import cached_property

import bittensor
from compute_horde_core.executor_class import ExecutorClass
from compute_horde_core.output_upload import OutputUpload
from compute_horde_core.streaming import StreamingDetails
from compute_horde_core.volume import Volume
from pydantic import TypeAdapter

from compute_horde.base.docker import DockerRunOptionsPreset
from compute_horde.executor_class import EXECUTOR_CLASS
from compute_horde.miner_client.base import (
    AbstractMinerClient,
    ErrorCallback,
    UnsupportedMessageReceived,
)
from compute_horde.protocol_messages import (
    GenericError,
    MinerToValidatorMessage,
    UnauthorizedError,
    V0AcceptJobRequest,
    V0DeclineJobRequest,
    V0ExecutionDoneRequest,
    V0ExecutorFailedRequest,
    V0ExecutorManifestRequest,
    V0ExecutorReadyRequest,
    V0InitialJobRequest,
    V0JobAcceptedReceiptRequest,
    V0JobFailedRequest,
    V0JobFinishedReceiptRequest,
    V0JobFinishedRequest,
    V0JobRequest,
    V0MachineSpecsRequest,
    V0StreamingJobNotReadyRequest,
    V0StreamingJobReadyRequest,
    V0VolumesReadyRequest,
    ValidatorAuthForMiner,
    ValidatorToMinerMessage,
)
from compute_horde.receipts.models import (
    JobAcceptedReceipt,
    JobFinishedReceipt,
    JobStartedReceipt,
)
from compute_horde.receipts.schemas import (
    JobAcceptedReceiptPayload,
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
)
from compute_horde.transport import (
    AbstractTransport,
    TransportConnectionError,
    WSTransport,
)
from compute_horde.utils import MachineSpecs, Timer, sign_blob

logger = logging.getLogger(__name__)


class OrganicMinerClient(AbstractMinerClient[MinerToValidatorMessage, ValidatorToMinerMessage]):
    """
    Miner client to run organic job on a miner.
    This client is used by validators to connect to prod miners.
    Others can use this client to connect to their locally running miners to run jobs.

    The usual usage of this miner client is as follows::

        client = OrganicMinerClient(...)
        async with client:
            await client.send_model(V0InitialJobRequest(...))
            msg = await client.miner_ready_or_declining_future
            # handle msg

            await client.send_model(V0JobRequest(...))
            msg = client.miner_finished_or_failed_future
            # handle msg

    Note that the waiting on the response futures should properly be handled with timeouts
    (with ``asyncio.timeout()``, ``asyncio.wait_for()`` etc.).
    """

    def __init__(
        self,
        miner_hotkey: str,
        miner_address: str,
        miner_port: int,
        job_uuid: str,
        my_keypair: bittensor.Keypair,
        transport: AbstractTransport | None = None,
    ) -> None:
        self.job_uuid = job_uuid

        self.miner_hotkey = miner_hotkey
        self.miner_address = miner_address
        self.miner_port = miner_port
        self.my_keypair = my_keypair

        loop = asyncio.get_running_loop()
        self.miner_manifest = loop.create_future()
        self.online_executor_count = 0

        # for waiting on miner responses
        self.miner_accepting_or_declining_future: asyncio.Future[
            V0AcceptJobRequest | V0DeclineJobRequest
        ] = loop.create_future()
        self.miner_accepting_or_declining_timestamp: int = 0

        self.executor_ready_or_failed_future: asyncio.Future[
            V0ExecutorReadyRequest | V0ExecutorFailedRequest
        ] = loop.create_future()
        self.executor_ready_or_failed_timestamp: int = 0

        self.streaming_job_ready_or_not_future: asyncio.Future[
            V0StreamingJobReadyRequest | V0StreamingJobNotReadyRequest
        ] = loop.create_future()
        self.streaming_job_ready_or_not_timestamp: int = 0

        self.volumes_ready_future: asyncio.Future[V0VolumesReadyRequest | V0JobFailedRequest] = (
            loop.create_future()
        )
        self.volumes_ready_timestamp: int = 0

        self.execution_done_future: asyncio.Future[V0ExecutionDoneRequest | V0JobFailedRequest] = (
            loop.create_future()
        )
        self.execution_done_timestamp: int = 0

        self.miner_finished_or_failed_future: asyncio.Future[
            V0JobFailedRequest | V0JobFinishedRequest
        ] = loop.create_future()
        self.miner_finished_or_failed_timestamp: int = 0

        self.miner_machine_specs: MachineSpecs | None = None

        name = f"{miner_hotkey}({miner_address}:{miner_port})"
        transport = transport or WSTransport(name, self.miner_url())
        super().__init__(name, transport)

    @cached_property
    def my_hotkey(self) -> str:
        return str(self.my_keypair.ss58_address)

    def miner_url(self) -> str:
        return (
            f"ws://{self.miner_address}:{self.miner_port}/v0.1/validator_interface/{self.my_hotkey}"
        )

    def parse_message(self, raw_msg: str | bytes) -> MinerToValidatorMessage:
        return TypeAdapter(MinerToValidatorMessage).validate_json(raw_msg)

    async def notify_generic_error(self, msg: GenericError) -> None:
        """This method is called when miner sends a generic error message"""

    async def notify_unauthorized_error(self, msg: UnauthorizedError) -> None:
        """This method is called when miner refuses the authentication message"""

    async def notify_receipt_failure(self, comment: str) -> None:
        """This method is called when sending receipts to miner fails"""

    async def notify_send_failure(self, msg: str) -> None:
        """This method is called when sending messages to miner fails"""

    async def notify_job_accepted(self, msg: V0AcceptJobRequest) -> None:
        """This method is called when miner sends job accepted message"""

    async def notify_executor_ready(self, msg: V0ExecutorReadyRequest) -> None:
        """This method is called when miner sends executor ready message"""

    async def notify_volumes_ready(self, msg: V0VolumesReadyRequest) -> None:
        """This method is called when miner sends executor ready message"""

    async def notify_execution_done(self, msg: V0ExecutionDoneRequest) -> None:
        """This method is called when miner sends execution done message"""

    async def notify_streaming_readiness(self, msg: V0StreamingJobReadyRequest) -> None:
        """This method is called when miner sends streaming ready message"""

    async def handle_manifest_request(self, msg: V0ExecutorManifestRequest) -> None:
        try:
            self.miner_manifest.set_result(msg.manifest)
        except asyncio.InvalidStateError:
            logger.warning(f"Received manifest from {msg} but future was already set")

    async def handle_machine_specs_request(self, msg: V0MachineSpecsRequest) -> None:
        self.miner_machine_specs = msg.specs

    async def handle_message(self, msg: MinerToValidatorMessage) -> None:
        logger.debug(f"Received message: {msg.message_type}")
        if isinstance(msg, GenericError):
            logger.warning(
                f"Received error message from miner {self.miner_name}: {msg.model_dump_json()}"
            )
            await self.notify_generic_error(msg)
            return
        elif isinstance(msg, UnauthorizedError):
            logger.error(f"Unauthorized in {self.miner_name}: {msg.code}, details: {msg.details}")
            await self.notify_unauthorized_error(msg)
            return
        elif isinstance(msg, V0ExecutorManifestRequest):
            await self.handle_manifest_request(msg)
            return

        if (received_job_uuid := getattr(msg, "job_uuid", self.job_uuid)) != self.job_uuid:
            logger.warning(
                f"Received msg from {self.miner_name} for a different job (expected job {self.job_uuid}, got job {received_job_uuid}): {msg}"
            )
            return

        if isinstance(msg, V0AcceptJobRequest | V0DeclineJobRequest):
            try:
                self.miner_accepting_or_declining_future.set_result(msg)
                self.miner_accepting_or_declining_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0ExecutorReadyRequest | V0ExecutorFailedRequest):
            try:
                self.executor_ready_or_failed_future.set_result(msg)
                self.executor_ready_or_failed_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0StreamingJobReadyRequest | V0StreamingJobNotReadyRequest):
            try:
                self.streaming_job_ready_or_not_future.set_result(msg)
                self.streaming_job_ready_or_not_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0VolumesReadyRequest):
            try:
                self.volumes_ready_future.set_result(msg)
                self.volumes_ready_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0ExecutionDoneRequest):
            try:
                self.execution_done_future.set_result(msg)
                self.execution_done_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0JobFinishedRequest):
            try:
                self.miner_finished_or_failed_future.set_result(msg)
                self.miner_finished_or_failed_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0JobFailedRequest):
            # Consider remaining stages as failed.
            try:
                self.volumes_ready_future.set_result(msg)
                self.volumes_ready_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                pass
            try:
                self.execution_done_future.set_result(msg)
                self.execution_done_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                pass
            try:
                self.miner_finished_or_failed_future.set_result(msg)
                self.miner_finished_or_failed_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                pass
        elif isinstance(msg, V0MachineSpecsRequest):
            await self.handle_machine_specs_request(msg)
        else:
            raise UnsupportedMessageReceived(msg)

    def generate_authentication_message(self) -> ValidatorAuthForMiner:
        msg = ValidatorAuthForMiner(
            validator_hotkey=self.my_hotkey,
            miner_hotkey=self.miner_hotkey,
            timestamp=int(time.time()),
            signature="",
        )
        msg.signature = sign_blob(self.my_keypair, msg.blob_for_signing())
        return msg

    def generate_job_started_receipt_message(
        self,
        executor_class: ExecutorClass,
        ttl: int,
    ) -> tuple[JobStartedReceiptPayload, str]:
        receipt_payload = JobStartedReceiptPayload(
            job_uuid=self.job_uuid,
            miner_hotkey=self.miner_hotkey,
            validator_hotkey=self.my_hotkey,
            timestamp=datetime.datetime.now(datetime.UTC),
            executor_class=executor_class,
            is_organic=True,
            ttl=ttl,
        )
        signature = sign_blob(self.my_keypair, receipt_payload.blob_for_signing())
        return receipt_payload, signature

    def generate_job_accepted_receipt_message(
        self,
        accepted_timestamp: float,
        ttl: int,
    ) -> V0JobAcceptedReceiptRequest:
        time_accepted = datetime.datetime.fromtimestamp(accepted_timestamp, datetime.UTC)
        receipt_payload = JobAcceptedReceiptPayload(
            job_uuid=self.job_uuid,
            miner_hotkey=self.miner_hotkey,
            validator_hotkey=self.my_hotkey,
            timestamp=datetime.datetime.now(datetime.UTC),
            time_accepted=time_accepted,
            ttl=ttl,
        )
        return V0JobAcceptedReceiptRequest(
            payload=receipt_payload,
            signature=sign_blob(self.my_keypair, receipt_payload.blob_for_signing()),
        )

    async def send_job_accepted_receipt_message(
        self,
        accepted_timestamp: float,
        ttl: int,
    ) -> None:
        try:
            receipt_message = self.generate_job_accepted_receipt_message(
                accepted_timestamp,
                ttl,
            )
            await self.send_model(receipt_message)
            await JobAcceptedReceipt.from_payload(
                receipt_message.payload,
                validator_signature=receipt_message.signature,
            ).asave()
            logger.debug(f"Sent job accepted receipt for {self.job_uuid}")
        except Exception as e:
            comment = f"Failed to send job accepted receipt to miner {self.miner_name} for job {self.job_uuid}: {e}"
            logger.warning(comment)
            await self.notify_receipt_failure(comment)

    def generate_job_finished_receipt_message(
        self,
        started_timestamp: float,
        time_took_seconds: float,
        score: float,
    ) -> V0JobFinishedReceiptRequest:
        time_started = datetime.datetime.fromtimestamp(started_timestamp, datetime.UTC)
        receipt_payload = JobFinishedReceiptPayload(
            job_uuid=self.job_uuid,
            miner_hotkey=self.miner_hotkey,
            validator_hotkey=self.my_hotkey,
            timestamp=datetime.datetime.now(datetime.UTC),
            time_started=time_started,
            time_took_us=int(time_took_seconds * 1_000_000),
            score_str=f"{score:.6f}",
        )
        return V0JobFinishedReceiptRequest(
            payload=receipt_payload,
            signature=sign_blob(self.my_keypair, receipt_payload.blob_for_signing()),
        )

    async def send_job_finished_receipt_message(
        self,
        started_timestamp: float,
        time_took_seconds: float,
        score: float,
    ) -> None:
        try:
            receipt_message = self.generate_job_finished_receipt_message(
                started_timestamp, time_took_seconds, score
            )
            await self.send_model(receipt_message)
            await JobFinishedReceipt.from_payload(
                receipt_message.payload,
                validator_signature=receipt_message.signature,
            ).asave()
            logger.debug(f"Sent job finished receipt for {self.job_uuid}")
        except Exception as e:
            comment = f"Failed to send job finished receipt to miner {self.miner_name} for job {self.job_uuid}: {e}"
            logger.warning(comment)
            await self.notify_receipt_failure(comment)

    async def send_initial_job_request(
        self,
        job_details: "OrganicJobDetails",
        receipt_payload: JobStartedReceiptPayload,
        receipt_signature: str,
    ) -> None:
        await self.send_model(
            V0InitialJobRequest(
                job_uuid=job_details.job_uuid,
                executor_class=job_details.executor_class,
                docker_image=job_details.docker_image,
                volume=job_details.volume,
                job_started_receipt_payload=receipt_payload,
                job_started_receipt_signature=receipt_signature,
                timeout_seconds=job_details.total_job_timeout,
                executor_timing=V0InitialJobRequest.ExecutorTimingDetails(
                    allowed_leeway=job_details.job_timing.allowed_leeway,
                    download_time_limit=job_details.job_timing.download_time_limit,
                    execution_time_limit=job_details.job_timing.execution_time_limit,
                    upload_time_limit=job_details.job_timing.upload_time_limit,
                    streaming_start_time_limit=job_details.job_timing.streaming_start_time_limit,
                )
                if job_details.job_timing is not None
                else None,
                streaming_details=job_details.streaming_details,
            ),
        )

    async def send_model(
        self, model: ValidatorToMinerMessage, error_event_callback: ErrorCallback | None = None
    ) -> None:
        if error_event_callback is None:
            error_event_callback = self.notify_send_failure

        await super().send_model(model, error_event_callback)

    async def connect(self) -> None:
        await super().connect()
        await self.transport.send(self.generate_authentication_message().model_dump_json())


class FailureReason(enum.Enum):
    MINER_CONNECTION_FAILED = enum.auto()
    INITIAL_RESPONSE_TIMED_OUT = enum.auto()
    EXECUTOR_READINESS_RESPONSE_TIMED_OUT = enum.auto()
    VOLUMES_TIMED_OUT = enum.auto()
    VOLUMES_FAILED = enum.auto()
    EXECUTION_TIMED_OUT = enum.auto()
    EXECUTION_FAILED = enum.auto()
    FINAL_RESPONSE_TIMED_OUT = enum.auto()
    JOB_DECLINED = enum.auto()
    EXECUTOR_FAILED = enum.auto()
    STREAMING_JOB_READY_TIMED_OUT = enum.auto()
    JOB_FAILED = enum.auto()
    STREAMING_FAILED = enum.auto()


class OrganicJobError(Exception):
    def __init__(self, reason: FailureReason, received: MinerToValidatorMessage | None = None):
        self.reason = reason
        self.received = received

    def __str__(self):
        s = f"Organic job failed, {self.reason=}"
        if self.received:
            s += f", received: {self.received_str()}"
        return s

    def __repr__(self):
        return f"{type(self).__name__}: {str(self)}"

    def received_str(self) -> str:
        if not self.received:
            return ""
        return self.received.model_dump_json()


@dataclass
class OrganicJobDetails:
    @dataclass
    class TimingDetails:
        allowed_leeway: int
        download_time_limit: int
        execution_time_limit: int
        upload_time_limit: int
        streaming_start_time_limit: int

        @property
        def total(self):
            return sum(asdict(self).values())

    job_uuid: str
    executor_class: ExecutorClass
    docker_image: str
    job_timing: TimingDetails | None = None
    docker_run_options_preset: DockerRunOptionsPreset = "nvidia_all"
    docker_run_cmd: list[str] = field(default_factory=list)
    total_job_timeout: int = 300  # Deprecated, use job_timing instead.
    volume: Volume | None = None
    output: OutputUpload | None = None
    artifacts_dir: str | None = None
    streaming_details: StreamingDetails | None = None


async def execute_organic_job_on_miner(
    client: OrganicMinerClient,
    job_details: OrganicJobDetails,
    reservation_time_limit: int,
    executor_startup_time_limit: int,
) -> tuple[str, str, dict[str, str], dict[str, str]]:  # stdout, stderr, artifacts, upload_results
    """
    Run an organic job. This is a simpler way to use OrganicMinerClient.
    :param client: the organic miner client
    :param job_details: details specific to the job that needs to be run
    :param reservation_time_limit: time for the miner to report reservation success (or decline the job)
    :param executor_startup_time_limit: time for executor to perform startup checks
    :return: standard out, standard error of the job container, artifacts and a dictionary mapping file names to HTTP upload responses
    """
    assert client.job_uuid == job_details.job_uuid

    timer = Timer()  # Only used for measurement, not used for timeouts.

    executor_spinup_time_limit = EXECUTOR_CLASS[job_details.executor_class].spin_up_time
    readiness_time_limit = executor_spinup_time_limit + executor_startup_time_limit

    # Note: If this is empty, we're in single-timeout mode.
    executor_timing = job_details.job_timing

    async with contextlib.AsyncExitStack() as exit_stack:
        try:
            await exit_stack.enter_async_context(client)
        except TransportConnectionError as exc:
            raise OrganicJobError(FailureReason.MINER_CONNECTION_FAILED) from exc

        ## STAGE: reservation
        # Miner should reserve an executor and respond with accept/reject quickly.
        # The receipt should only be good enough for the reservation time.
        # Miner will receive a "job accepted" receipt with a longer TTL as soon as we get a reservation.
        receipt_payload, receipt_signature = client.generate_job_started_receipt_message(
            executor_class=job_details.executor_class,
            ttl=reservation_time_limit,
        )
        await client.send_initial_job_request(job_details, receipt_payload, receipt_signature)
        logger.debug("Sent initial job request")
        await JobStartedReceipt.from_payload(
            receipt_payload,
            validator_signature=receipt_signature,
        ).asave()

        try:
            try:
                logger.debug(f"Waiting for initial response for {reservation_time_limit:.2f}s")
                initial_response = await asyncio.wait_for(
                    client.miner_accepting_or_declining_future,
                    timeout=reservation_time_limit,
                )
            except TimeoutError as exc:
                raise OrganicJobError(FailureReason.INITIAL_RESPONSE_TIMED_OUT) from exc
            if isinstance(initial_response, V0DeclineJobRequest):
                raise OrganicJobError(FailureReason.JOB_DECLINED, initial_response)

            await client.notify_job_accepted(initial_response)

            ## STAGE: Spinup
            # Validator waits for the executor to start up and report readiness.
            # This includes executor startup checks.
            if executor_timing:
                job_accepted_receipt_ttl = readiness_time_limit + executor_timing.total
            else:
                job_accepted_receipt_ttl = readiness_time_limit + job_details.total_job_timeout
            try:
                await client.send_job_accepted_receipt_message(
                    accepted_timestamp=time.time(),
                    ttl=job_accepted_receipt_ttl,
                )
                executor_readiness_response = await asyncio.wait_for(
                    client.executor_ready_or_failed_future,
                    timeout=readiness_time_limit,
                )
            except TimeoutError as exc:
                raise OrganicJobError(FailureReason.EXECUTOR_READINESS_RESPONSE_TIMED_OUT) from exc
            if isinstance(executor_readiness_response, V0ExecutorFailedRequest):
                raise OrganicJobError(FailureReason.EXECUTOR_FAILED, executor_readiness_response)

            await client.notify_executor_ready(executor_readiness_response)

            if executor_timing is not None:
                # For fine-grained timeouts, the deadline will be extended at the start of each stage.
                logger.debug(f"Starting deadline with {executor_timing.allowed_leeway}s leeway")
                deadline = Timer(executor_timing.allowed_leeway)
            else:
                # For single-timeout mode, the deadline is set once here.
                logger.debug(
                    f"Starting deadline with {job_details.total_job_timeout}s total timeout"
                )
                deadline = Timer(job_details.total_job_timeout)

            await client.send_model(
                V0JobRequest(
                    job_uuid=job_details.job_uuid,
                    executor_class=job_details.executor_class,
                    docker_image=job_details.docker_image,
                    docker_run_options_preset=job_details.docker_run_options_preset,
                    docker_run_cmd=job_details.docker_run_cmd,
                    volume=None,  # Was sent in the initial request
                    output_upload=job_details.output,
                    artifacts_dir=job_details.artifacts_dir,
                )
            )

            ## STAGE: Volume download
            try:
                if executor_timing:
                    logger.debug(
                        f"Extending deadline by download_time_limit: +{executor_timing.download_time_limit}s"
                    )
                    deadline.extend_timeout(executor_timing.download_time_limit)
                logger.debug(
                    f"Waiting for volume download (time left: {deadline.time_left():.2f}s)"
                )
                volumes_ready_response = await asyncio.wait_for(
                    client.volumes_ready_future,
                    timeout=deadline.time_left(),
                )
                if isinstance(volumes_ready_response, V0JobFailedRequest):
                    raise OrganicJobError(FailureReason.VOLUMES_FAILED, volumes_ready_response)
                logger.debug(f"Volume download done with {deadline.time_left():.2f}s left")
            except TimeoutError as exc:
                raise OrganicJobError(FailureReason.VOLUMES_TIMED_OUT) from exc
            await client.notify_volumes_ready(volumes_ready_response)

            ## STAGE: Start streaming
            if job_details.streaming_details:
                try:
                    if executor_timing:
                        logger.debug(
                            f"Extending deadline by streaming_start_time_limit: +{executor_timing.streaming_start_time_limit}s"
                        )
                        deadline.extend_timeout(executor_timing.streaming_start_time_limit)
                    logger.debug(f"Waiting for streaming (time left: {deadline.time_left():.2f}s)")
                    streaming_response = await asyncio.wait_for(
                        client.streaming_job_ready_or_not_future,
                        timeout=deadline.time_left(),
                    )
                except TimeoutError as exc:
                    raise OrganicJobError(FailureReason.STREAMING_JOB_READY_TIMED_OUT) from exc
                if isinstance(streaming_response, V0StreamingJobNotReadyRequest):
                    raise OrganicJobError(FailureReason.STREAMING_FAILED, streaming_response)

                await client.notify_streaming_readiness(streaming_response)

            ## STAGE: execution
            try:
                if executor_timing:
                    logger.debug(
                        f"Extending deadline by execution_time_limit: +{executor_timing.execution_time_limit}s"
                    )
                    deadline.extend_timeout(executor_timing.execution_time_limit)
                logger.debug(f"Waiting for execution (time left: {deadline.time_left():.2f}s)")
                execution_done_response = await asyncio.wait_for(
                    client.execution_done_future,
                    timeout=deadline.time_left(),
                )
                if isinstance(execution_done_response, V0JobFailedRequest):
                    raise OrganicJobError(FailureReason.EXECUTION_FAILED, execution_done_response)
                logger.debug(f"Execution done with {deadline.time_left():.2f}s left")
            except TimeoutError as exc:
                raise OrganicJobError(FailureReason.EXECUTION_TIMED_OUT) from exc
            await client.notify_execution_done(execution_done_response)

            ## STAGE: upload
            try:
                if executor_timing:
                    logger.debug(
                        f"Extending deadline by upload_time_limit: +{executor_timing.upload_time_limit}s"
                    )
                    deadline.extend_timeout(executor_timing.upload_time_limit)
                logger.debug(f"Waiting for upload (time left: {deadline.time_left():.2f}s)")
                final_response = await asyncio.wait_for(
                    client.miner_finished_or_failed_future,
                    timeout=deadline.time_left(),
                )
                if isinstance(final_response, V0JobFailedRequest):
                    raise OrganicJobError(FailureReason.JOB_FAILED, final_response)
                logger.debug(f"Upload done with {deadline.time_left():.2f}s left")

                logger.info(f"Job finished in time with {deadline.time_left():.2f}s left")
                await client.send_job_finished_receipt_message(
                    started_timestamp=timer.start_time.timestamp(),
                    time_took_seconds=timer.passed_time(),
                    score=0,  # no score for organic jobs (at least right now)
                )

                return (
                    final_response.docker_process_stdout,
                    final_response.docker_process_stderr,
                    final_response.artifacts or {},
                    final_response.upload_results or {},
                )
            except TimeoutError as exc:
                raise OrganicJobError(FailureReason.FINAL_RESPONSE_TIMED_OUT) from exc

        except Exception as e:
            logger.warning(f"Job failed with {type(e).__name__}: {e}")
            await client.send_job_finished_receipt_message(
                started_timestamp=timer.start_time.timestamp(),
                time_took_seconds=timer.passed_time(),
                score=0,
            )
            raise

    raise Exception("Organic job flow ended with no result")



from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured

REQUIRED_SETTINGS = [
    "COMPUTE_HORDE_VALIDATOR_MODEL",
    "COMPUTE_HORDE_VALIDATOR_KEY_FIELD",
    "COMPUTE_HORDE_VALIDATOR_ACTIVE_FIELD",
    # COMPUTE_HORDE_VALIDATOR_DEBUG_FIELD - optional
    "BITTENSOR_NETUID",
    "BITTENSOR_NETWORK",
]


class ValidatorsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "compute_horde.validators"
    verbose_name = "validators"

    def ready(self):
        from django.conf import settings

        # Check if required settings are defined
        missing_settings = [
            setting for setting in REQUIRED_SETTINGS if not hasattr(settings, setting)
        ]

        if missing_settings:
            raise ImproperlyConfigured(
                f"The settings: {', '.join(missing_settings)} are required in your settings.py "
                "for 'compute_horde.validators' to function properly. Please define them."
            )



from django.contrib import admin  # noqa

from compute_horde.base.admin import ReadOnlyAdminMixin
from compute_horde.receipts.models import JobFinishedReceipt, JobStartedReceipt, JobAcceptedReceipt


class JobStartedReceiptsReadOnlyAdmin(admin.ModelAdmin, ReadOnlyAdminMixin):
    list_display = [
        "job_uuid",
        "miner_hotkey",
        "validator_hotkey",
        "timestamp",
        "executor_class",
        "ttl",
    ]
    ordering = ["-timestamp"]


class JobAcceptedReceiptsReadOnlyAdmin(admin.ModelAdmin, ReadOnlyAdminMixin):
    list_display = [
        "job_uuid",
        "miner_hotkey",
        "validator_hotkey",
        "timestamp",
        "time_accepted",
        "ttl",
    ]
    ordering = ["-timestamp"]


class JobFinishedReceiptsReadOnlyAdmin(admin.ModelAdmin, ReadOnlyAdminMixin):
    list_display = [
        "job_uuid",
        "miner_hotkey",
        "validator_hotkey",
        "timestamp",
        "score",
        "time_started",
        "time_took",
    ]
    ordering = ["-timestamp"]


admin.site.register(JobStartedReceipt, admin_class=JobStartedReceiptsReadOnlyAdmin)
admin.site.register(JobAcceptedReceipt, admin_class=JobAcceptedReceiptsReadOnlyAdmin)
admin.site.register(JobFinishedReceipt, admin_class=JobFinishedReceiptsReadOnlyAdmin)
